"""Defines `meltano elt` command."""

from __future__ import annotations

import logging
import platform
import typing as t
from contextlib import asynccontextmanager, nullcontext, suppress
from datetime import datetime, timezone

import click
from structlog import stdlib as structlog_stdlib

from meltano.cli.params import pass_project
from meltano.cli.utils import CliEnvironmentBehavior, CliError, PartialInstrumentedCmd
from meltano.core.db import project_engine
from meltano.core.elt_context import ELTContextBuilder
from meltano.core.job import Job, JobFinder
from meltano.core.job.stale_job_failer import fail_stale_jobs
from meltano.core.logging import JobLoggingService, OutputLogger
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.runner import RunnerError
from meltano.core.runner.dbt import DbtRunner
from meltano.core.runner.singer import SingerRunner
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext
from meltano.core.utils import run_async

if t.TYPE_CHECKING:
    import structlog

    from meltano.core.project import Project
    from meltano.core.tracking import Tracker

DUMPABLES = {
    "catalog": (PluginType.EXTRACTORS, "catalog"),
    "state": (PluginType.EXTRACTORS, "state"),
    "extractor-config": (PluginType.EXTRACTORS, "config"),
    "loader-config": (PluginType.LOADERS, "config"),
}

logger = structlog_stdlib.get_logger(__name__)


class ELOptions:
    """Namespace for all options shared by the `el` and `elt` commands."""

    extractor = click.argument("extractor")
    loader = click.argument("loader")
    transform = click.option(
        "--transform",
        type=click.Choice(["skip", "only", "run"]),
    )
    dry = click.option("--dry", help="Do not actually run.", is_flag=True)
    full_refresh = click.option(
        "--full-refresh",
        help="Perform a full refresh (ignore state left behind by any previous runs).",
        is_flag=True,
    )
    select = click.option(
        "--select",
        "-s",
        help="Select only these specific entities for extraction.",
        multiple=True,
        default=[],
    )
    exclude = click.option(
        "--exclude",
        "-e",
        help="Exclude these specific entities from extraction.",
        multiple=True,
        default=[],
    )
    catalog = click.option("--catalog", help="Extractor catalog file.")
    state = click.option("--state", help="Extractor state file.")
    dump = click.option(
        "--dump",
        type=click.Choice(DUMPABLES.keys()),
        help="Dump content of pipeline-specific generated file.",
    )
    state_id = click.option(
        "--state-id",
        envvar="MELTANO_STATE_ID",
        help="A custom string to identify the job.",
    )
    force = click.option(
        "--force",
        "-f",
        help=(
            "Force a new run even when a pipeline with the same state ID is "
            "already running."
        ),
        is_flag=True,
    )
    merge_state = click.option(
        "--merge-state",
        is_flag=True,
        help="Merges state with that of previous runs.",
    )


@click.command(
    cls=PartialInstrumentedCmd,
    short_help="Run an EL pipeline to Extract and Load data.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_use_default,
)
@ELOptions.extractor
@ELOptions.loader
@ELOptions.dry
@ELOptions.full_refresh
@ELOptions.select
@ELOptions.exclude
@ELOptions.catalog
@ELOptions.state
@ELOptions.dump
@ELOptions.state_id
@ELOptions.force
@ELOptions.merge_state
@click.pass_context
@pass_project(migrate=True)
@run_async
async def el(  # WPS408
    project: Project,
    ctx: click.Context,
    extractor: str,
    loader: str,
    dry: bool,
    full_refresh: bool,
    select: list[str],
    exclude: list[str],
    catalog: str,
    state: str,
    dump: str,
    state_id: str,
    force: bool,
    merge_state: bool,
):
    """
    Run an EL pipeline to Extract and Load data.

    meltano el '<extractor_name>' '<loader_name>'

    extractor_name: extractor to be used in this pipeline.
    loader_name: loader to be used in this pipeline.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#el
    """
    await _run_el_command(
        project,
        ctx,
        extractor,
        loader,
        None,
        dry,
        full_refresh,
        select,
        exclude,
        catalog,
        state,
        dump,
        state_id,
        force,
        merge_state,
    )


@click.command(
    cls=PartialInstrumentedCmd,
    short_help="Run an ELT pipeline to Extract, Load, and Transform data.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_use_default,
)
@ELOptions.extractor
@ELOptions.loader
@ELOptions.transform
@ELOptions.dry
@ELOptions.full_refresh
@ELOptions.select
@ELOptions.exclude
@ELOptions.catalog
@ELOptions.state
@ELOptions.dump
@ELOptions.state_id
@ELOptions.force
@ELOptions.merge_state
@click.pass_context
@pass_project(migrate=True)
@run_async
async def elt(  # WPS408
    project: Project,
    ctx: click.Context,
    extractor: str,
    loader: str,
    transform: str,
    dry: bool,
    full_refresh: bool,
    select: list[str],
    exclude: list[str],
    catalog: str,
    state: str,
    dump: str,
    state_id: str,
    force: bool,
    merge_state: bool,
):
    """
    Run an ELT pipeline to Extract, Load, and Transform data.

    meltano elt '<extractor_name>' '<loader_name>'

    extractor_name: extractor to be used in this pipeline.
    loader_name: loader to be used in this pipeline.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#elt
    """
    logger.warning("The `elt` command is deprecated in favor of `el`")
    await _run_el_command(
        project,
        ctx,
        extractor,
        loader,
        transform,
        dry,
        full_refresh,
        select,
        exclude,
        catalog,
        state,
        dump,
        state_id,
        force,
        merge_state,
    )


async def _run_el_command(
    project: Project,
    ctx: click.Context,
    extractor: str,
    loader: str,
    transform: str | None,
    dry: bool,
    full_refresh: bool,
    select: list[str],
    exclude: list[str],
    catalog: str,
    state: str,
    dump: str,
    state_id: str,
    force: bool,
    merge_state: bool,
):
    if platform.system() == "Windows":
        raise CliError(
            "ELT command not supported on Windows. Please use the run command "  # noqa: EM101
            "as documented here: "
            "https://docs.meltano.com/reference/command-line-interface#run",
        )

    tracker: Tracker = ctx.obj["tracker"]

    # We no longer set a default choice for transform, so that we can detect
    # explicit usages of the `--transform` option if transform is `None` we
    # still need manually default to skip after firing the tracking event above
    # TODO: Remove the transform option entirely in a future release
    transform = transform or "skip"

    select_filter = [*select, *(f"!{entity}" for entity in exclude)]

    job = Job(
        job_name=state_id
        or (
            f'{datetime.now(timezone.utc).strftime("%Y-%m-%dT%H%M%S")}'
            f"--{extractor}--{loader}"
        ),
    )
    _, Session = project_engine(project)  # noqa: N806
    session = Session()
    try:
        context_builder = _elt_context_builder(
            project,
            job,
            session,
            extractor,
            loader,
            transform,
            dry_run=dry,
            full_refresh=full_refresh,
            select_filter=select_filter,
            catalog=catalog,
            state=state,
            merge_state=merge_state,
        )

        if dump:
            await dump_file(context_builder, dump)
        else:
            await _run_job(tracker, project, job, session, context_builder, force=force)
    except Exception as err:
        tracker.track_command_event(CliEvent.failed)
        raise err
    finally:
        session.close()

    tracker.track_command_event(CliEvent.completed)


def _elt_context_builder(
    project: Project,
    job,
    session,
    extractor,
    loader,
    transform,
    dry_run=False,
    full_refresh=False,
    select_filter=None,
    catalog=None,
    state=None,
    merge_state=False,
):
    select_filter = select_filter or []
    transform_name = None
    if transform != "skip":
        transform_name = _find_transform_for_extractor(project, extractor)

    return (
        ELTContextBuilder(project)
        .with_session(session)
        .with_job(job)
        .with_extractor(extractor)
        .with_loader(loader)
        .with_transform(transform_name or transform)
        .with_dry_run(dry_run)
        .with_only_transform(transform == "only")
        .with_full_refresh(full_refresh)
        .with_select_filter(select_filter)
        .with_catalog(catalog)
        .with_state(state)
        .with_merge_state(merge_state)
    )


async def dump_file(context_builder, dumpable):
    """Dump the given dumpable for this pipeline."""
    elt_context = context_builder.context()

    try:
        plugin_type, file_id = DUMPABLES[dumpable]
        invoker = elt_context.invoker_for(plugin_type)

        async with invoker.prepared(elt_context.session):
            content = await invoker.dump(file_id)

        click.echo(content)
    except FileNotFoundError as err:
        raise CliError(f"Could not find {dumpable} file for this pipeline") from err  # noqa: EM102
    except Exception as err:
        raise CliError(f"Could not dump {dumpable}: {err}") from err  # noqa: EM102


async def _run_job(tracker, project, job, session, context_builder, force=False):
    fail_stale_jobs(session, job.job_name)

    if not force and (existing := JobFinder(job.job_name).latest_running(session)):
        raise CliError(
            f"Another '{job.job_name}' pipeline is already running which "  # noqa: EM102
            f"started at {existing.started_at}. To ignore this check use "
            "the '--force' option.",
        )

    async with job.run(session):
        job_logging_service = JobLoggingService(project)
        log_file = job_logging_service.generate_log_name(job.job_name, job.run_id)

        output_logger = OutputLogger(log_file)
        context_builder.set_base_output_logger(output_logger)

        log = logger.bind(name="meltano", run_id=str(job.run_id), state_id=job.job_name)

        await _run_elt(tracker, log, context_builder, output_logger)


@asynccontextmanager
async def _redirect_output(log, output_logger):
    meltano_stdout = output_logger.out(
        "meltano",
        log.bind(stdio="stdout", cmd_type="elt"),
    )
    meltano_stderr = output_logger.out(
        "meltano",
        log.bind(stdio="stderr", cmd_type="elt"),
    )

    with meltano_stdout.redirect_logging(ignore_errors=(CliError,)):
        async with meltano_stdout.redirect_stdout(), meltano_stderr.redirect_stderr():  # noqa: WPS316, E501
            try:
                yield
            except CliError as err:
                err.print()
                raise


async def _run_elt(
    tracker: Tracker,
    log: structlog.BoundLogger,
    context_builder: ELTContextBuilder,
    output_logger: OutputLogger,
):
    async with _redirect_output(log, output_logger):
        try:
            elt_context = context_builder.context()
            tracker.add_contexts(PluginsTrackingContext.from_elt_context(elt_context))
            tracker.track_command_event(CliEvent.inflight)

            if elt_context.only_transform:
                log.info("Extract & load skipped.")
            else:
                await _run_extract_load(log, elt_context, output_logger)

            if elt_context.transformer:
                log.warning(
                    "The option to run a transformation is deprecated and will be "
                    "removed in a future version.",
                )
                await _run_transform(log, elt_context, output_logger)
            else:
                log.info("Transformation skipped.")
        except RunnerError as err:
            raise CliError(
                f"ELT could not be completed: {err}.\nFor more detailed log "  # noqa: EM102
                "messages re-run the command using 'meltano "
                "--log-level=debug ...' CLI flag.\nNote that you can also "
                f"check the generated log file at '{output_logger.file}'.\n"
                "For more information on debugging and logging: "
                "https://docs.meltano.com/reference/command-line-interface#debugging",
            ) from err


async def _run_extract_load(  # noqa: WPS210, WPS231
    log,
    elt_context,
    output_logger,
    **kwargs,
):
    extractor = elt_context.extractor.name
    loader = elt_context.loader.name

    stderr_log = logger.bind(
        run_id=str(elt_context.job.run_id),
        state_id=elt_context.job.job_name,
        stdio="stderr",
    )

    extractor_log = output_logger.out(extractor, stderr_log.bind(cmd_type="extractor"))
    loader_log = output_logger.out(loader, stderr_log.bind(cmd_type="loader"))

    extractor_out_writer_ctxmgr = nullcontext
    loader_out_writer_ctxmgr = nullcontext
    if logger.getEffectiveLevel() == logging.DEBUG:
        stdout_log = logger.bind(
            run_id=str(elt_context.job.run_id),
            state_id=elt_context.job.job_name,
            stdio="stdout",
        )
        extractor_out = output_logger.out(
            f"{extractor} (out)",
            stdout_log.bind(cmd_type="extractor"),
            logging.DEBUG,
        )
        loader_out = output_logger.out(
            f"{loader} (out)",
            stdout_log.bind(cmd_type="loader"),
            logging.DEBUG,
        )

        extractor_out_writer_ctxmgr = extractor_out.line_writer
        loader_out_writer_ctxmgr = loader_out.line_writer

    log.info(
        "Running extract & load...",
    )

    singer_runner = SingerRunner(elt_context)
    try:
        # Once Python 3.9 support has been dropped, consolidate these
        # with-statements by using parentheses.
        with extractor_log.line_writer() as extractor_log_writer:  # noqa: SIM117
            with loader_log.line_writer() as loader_log_writer:
                with extractor_out_writer_ctxmgr() as extractor_out_writer:
                    with loader_out_writer_ctxmgr() as loader_out_writer:
                        await singer_runner.run(
                            **kwargs,
                            extractor_log=extractor_log_writer,
                            loader_log=loader_log_writer,
                            extractor_out=extractor_out_writer,
                            loader_out=loader_out_writer,
                        )
    except RunnerError as err:
        with suppress(KeyError):
            code = err.exitcodes[PluginType.EXTRACTORS]
            message = extractor_log.last_line.rstrip() or "(see above)"
            log.error("Extraction failed", code=code, message=message)
        with suppress(KeyError):
            code = err.exitcodes[PluginType.LOADERS]
            message = loader_log.last_line.rstrip() or "(see above)"
            log.error("Loading failed", code=code, message=message)
        raise

    log.info("Extract & load complete!")


async def _run_transform(log, elt_context, output_logger, **kwargs):
    stderr_log = logger.bind(
        run_id=str(elt_context.job.run_id),
        state_id=elt_context.job.job_name,
        stdio="stderr",
        cmd_type="transformer",
    )

    transformer_log = output_logger.out(elt_context.transformer.name, stderr_log)

    log.info("Running transformation...")

    dbt_runner = DbtRunner(elt_context)
    try:
        with transformer_log.line_writer() as transformer_log_writer:
            await dbt_runner.run(**kwargs, log=transformer_log_writer)
    except RunnerError as err:
        with suppress(KeyError):
            code = err.exitcodes[PluginType.TRANSFORMERS]
            message = transformer_log.last_line.rstrip() or "(see above)"
            log.error("Transformation failed", code=code, message=message)
        raise

    log.info("Transformation complete!")


def _find_extractor(project: Project, extractor_name: str):
    try:
        return project.plugins.locked_definition_service.find_definition(
            PluginType.EXTRACTORS,
            extractor_name,
        )
    except PluginNotFoundError:
        return project.hub_service.find_definition(
            PluginType.EXTRACTORS,
            extractor_name,
        )


def _find_transform_for_extractor(
    project: Project,
    extractor_name: str,
):
    try:
        # Check if there is a default transform for this extractor
        transform_plugin_def = project.plugins.find_plugin_by_namespace(
            PluginType.TRANSFORMS,
            _find_extractor(project, extractor_name).namespace,
        )

        # Check if the transform has been added to the project
        transform_plugin = project.plugins.get_plugin(transform_plugin_def)

        return transform_plugin.name
    except PluginNotFoundError:
        return None



"""New Project Initialization Service."""

from __future__ import annotations

import contextlib
import os
import uuid
from pathlib import Path

import click

from meltano.core.cli_messages import GREETING
from meltano.core.db import project_engine
from meltano.core.plugin.meltano_file import MeltanoFilePlugin
from meltano.core.project import Project
from meltano.core.project_settings_service import SettingValueStore

PROJECT_FILENAME = "meltano.yml"


class ProjectInitServiceError(Exception):
    """Project Initialization Service Exception."""


class ProjectInitService:
    """New Project Initialization Service."""

    def __init__(self, project_directory: os.PathLike):
        """Create a new ProjectInitService instance.

        Args:
            project_directory: The directory path to create the project at
        """
        self.project_directory = Path(project_directory)

        with contextlib.suppress(ValueError):
            self.project_directory = self.project_directory.relative_to(Path.cwd())

    def init(  # noqa: C901
        self,
        *,
        activate: bool = True,
        force: bool = False,
    ) -> Project:
        """Initialise Meltano Project.

        Args:
            activate: Activate newly created project
            force: Whether to overwrite `meltano.yml` in the existing directory.

        Returns:
            A new Project instance

        Raises:
            ProjectInitServiceError: Directory already exists
        """
        try:
            self.project_directory.mkdir()
        except FileExistsError as ex:
            if (
                os.path.exists(os.path.join(self.project_directory, PROJECT_FILENAME))
                and not force
            ):
                msg = (
                    "A `meltano.yml` file already exists in the target directory. "
                    "Use `--force` to overwrite it."
                )
                raise ProjectInitServiceError(msg) from ex
        except PermissionError as ex:
            raise ProjectInitServiceError(
                f"Permission denied to create '{self.project_directory}'.",  # noqa: EM102
            ) from ex
        except Exception as ex:
            raise ProjectInitServiceError(
                f"Could not create directory '{self.project_directory}'. {ex}",  # noqa: EM102
            ) from ex

        project = Project(self.project_directory)

        self.create_dot_meltano_dir(project)
        self.create_files(project)

        project.settings.set(
            "project_id",
            str(uuid.uuid4()),
            store=SettingValueStore.MELTANO_YML,
        )
        self.set_send_anonymous_usage_stats(project)
        if activate:
            Project.activate(project)

        self.create_system_database(project)

        return project

    def create_dot_meltano_dir(self, project: Project):
        """Create .meltano directory.

        Args:
            project: Meltano project context
        """
        # explicitly create the .meltano directory if it doesn't exist
        click.secho("Creating .meltano folder", fg="blue")
        os.makedirs(project.meltano_dir(), exist_ok=True)
        click.secho("created", fg="blue", nl=False)
        click.echo(f" .meltano in {project.sys_dir_root}")

    def create_files(self, project: Project):
        """Create project files.

        Args:
            project: Meltano project context
        """
        click.secho("Creating project files...", fg="blue")

        if project.root != Path.cwd():
            click.echo(f"  {self.project_directory}/")

        plugin = MeltanoFilePlugin()

        expected_files = plugin.files_to_create(project, [])
        created_files = plugin.create_files(project)
        for path in expected_files:
            if path in created_files:
                click.secho("   |--", fg="blue", nl=False)
                click.echo(f" {path}")
            else:
                click.secho("   |--", fg="yellow", nl=False)
                click.echo(f" {path} (skipped)")

    def set_send_anonymous_usage_stats(self, project: Project):
        """Set Anonymous Usage Stats flag.

        Args:
            project: The Meltano project the setting is being set on.
        """
        # If set to false store explicitly in `meltano.yml`
        if not project.settings.get("send_anonymous_usage_stats"):
            project.settings.set(
                "send_anonymous_usage_stats",
                project.settings.get("send_anonymous_usage_stats"),
                store=SettingValueStore.MELTANO_YML,
            )

    def create_system_database(self, project: Project):
        """Create Meltano System DB.

        Args:
            project: Meltano project context

        Raises:
            ProjectInitServiceError: Database initialization failed
        """
        click.secho("Creating system database...", fg="blue", nl=False)

        # register the system database connection
        engine, _ = project_engine(project, default=True)

        from meltano.core.migration_service import MigrationError, MigrationService

        try:
            migration_service = MigrationService(engine)
            migration_service.upgrade(silent=True)
            click.secho("  Done!", fg="blue")
        except MigrationError as err:
            raise ProjectInitServiceError(str(err)) from err

    def echo_instructions(self, project: Project):
        """Echo Next Steps to Click CLI.

        Args:
            project: Meltano project context
        """
        click.secho(GREETING, nl=False)
        click.echo("Your project has been created!\n")

        click.echo("Meltano Environments initialized with ", nl=False)
        click.secho("dev", fg="bright_green", nl=False)
        click.echo(", ", nl=False)
        click.secho("staging", fg="bright_yellow", nl=False)
        click.echo(", and ", nl=False)
        click.secho("prod", fg="bright_red", nl=False)
        click.echo(".")
        click.echo("To learn more about Environments visit: ", nl=False)
        click.secho(
            "https://docs.meltano.com/concepts/environments",
            fg="cyan",
        )

        click.echo("\nNext steps:")

        if project.root != Path.cwd():
            click.secho("  cd ", nl=False)
            click.secho(self.project_directory, fg="magenta")

        click.echo("  Visit ", nl=False)
        click.secho(
            "https://docs.meltano.com/getting-started/part1",
            fg="cyan",
            nl=False,
        )
        click.echo(" to learn where to go from here")



"""Meltano schedule definition."""

from __future__ import annotations

import typing as t

from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical
from meltano.core.job import Job as StateJob
from meltano.core.job import JobFinder as StateJobFinder

if t.TYPE_CHECKING:
    import datetime

CRON_INTERVALS: dict[str, str | None] = {
    "@once": None,
    "@manual": None,
    "@none": None,
    "@hourly": "0 * * * *",
    "@daily": "0 0 * * *",
    "@weekly": "0 0 * * 0",
    "@monthly": "0 0 1 * *",
    "@yearly": "0 0 1 1 *",
}


class Schedule(NameEq, Canonical):  # noqa: WPS230
    """A schedule is an elt command or a job configured to run at a certain interval."""

    def __init__(
        self,
        *,
        name: str,
        extractor: str | None = None,
        loader: str | None = None,
        transform: str | None = None,
        interval: str | None,
        start_date: datetime.datetime | None = None,
        job: str | None = None,
        env: dict[str, str] | None = None,
    ):
        """Initialize a Schedule.

        Args:
            name: The name of the schedule.
            extractor: The name of the extractor.
            loader: The name of the loader.
            transform: The transform statement (eg: skip, only, run)
            interval: The interval of the schedule.
            start_date: The start date of the schedule.
            job: The name of the job.
            env: The env for this schedule.
        """
        if not env:
            env = {}

        super().__init__()

        # Attributes will be listed in meltano.yml in this order:
        self.name = name

        self.interval = interval
        self.job = job
        self.env = env

        if self.job:
            self.extractor = None
            self.loader = None
            self.transform = None
            self.start_date = None
        else:
            self.extractor = extractor
            self.loader = loader
            self.transform = transform
            self.start_date = start_date

    @property
    def cron_interval(self) -> str | None:
        """Return the explicit cron interval expression for a cron alias.

        Returns:
            The cron expression.
        """
        if self.interval:
            return CRON_INTERVALS.get(self.interval, self.interval)
        return None

    @property
    def elt_schedule(self) -> bool:
        """Return whether this schedule is an elt schedule.

        Returns:
            True if this schedule is an elt schedule.
        """
        return not self.job

    @property
    def elt_args(self) -> list[str]:
        """Return the list of arguments to pass to the elt command.

        Only valid if the schedule is an elt schedule.

        Returns:
            The list of arguments to pass to the elt command.

        Raises:
            NotImplementedError: If the schedule is a job.
        """
        if self.job:
            raise NotImplementedError
        return [
            t.cast(str, self.extractor),
            t.cast(str, self.loader),
            f"--transform={self.transform}",
            f"--state-id={self.name}",
        ]

    def last_successful_run(self, session) -> StateJob:
        """Return the last successful run for this schedule.

        Args:
            session: The database session.

        Returns:
            The last successful run for this schedule.

        Raises:
            NotImplementedError: If the schedule is a job.
        """
        if self.job:
            raise NotImplementedError(
                "Can't obtain last successful State(Job) for schedule job.",  # noqa: EM101
            )
        return StateJobFinder(self.name).latest_success(session)



from __future__ import annotations

import json
import os
import platform
import shutil
import typing as t
from pathlib import Path

import mock
import pytest
import yaml

from asserts import assert_cli_runner
from fixtures.cli import plugins_dir
from meltano.cli import cli
from meltano.cli.utils import CliError
from meltano.core.plugin import PluginRef, PluginType, Variant
from meltano.core.plugin.error import InvalidPluginDefinitionError, PluginNotFoundError
from meltano.core.plugin_install_service import PluginInstallReason

if t.TYPE_CHECKING:
    from click.testing import CliRunner

    from meltano.core.plugin.project_plugin import ProjectPlugin
    from meltano.core.project import Project
    from meltano.core.project_init_service import ProjectInitService

plugin_ref = plugins_dir / "extractors" / "tap-custom" / "test.yml"
fails_on_windows = pytest.mark.xfail(
    platform.system() == "Windows",
    reason="Fails on Windows: https://github.com/meltano/meltano/issues/3444",
)


class TestCliAdd:
    @pytest.fixture()
    def reset_project_context(
        self,
        project: Project,
        project_init_service: ProjectInitService,
    ):
        shutil.rmtree(".", ignore_errors=True)
        project_init_service.create_files(project)

        project.refresh()

        for plugin_type in PluginType:
            project.meltano.plugins[plugin_type].clear()

    @pytest.mark.order(0)
    @pytest.mark.parametrize(
        ("plugin_type", "plugin_name", "default_variant", "required_plugin_refs"),
        (
            (PluginType.EXTRACTORS, "tap-carbon-intensity", "meltano", []),
            (PluginType.LOADERS, "target-sqlite", "meltanolabs", []),
            (PluginType.TRANSFORMS, "tap-carbon-intensity", "meltano", []),
            (
                PluginType.ORCHESTRATORS,
                "airflow",
                Variant.ORIGINAL_NAME,
                [PluginRef(PluginType.FILES, "airflow")],
            ),
        ),
        ids=[
            "single-extractor",
            "single-loader",
            "transform-and-related",
            "orchestrator-and-required",
        ],
    )
    def test_add(
        self,
        plugin_type,
        plugin_name,
        default_variant,
        required_plugin_refs,
        project: Project,
        cli_runner,
    ):
        # ensure the plugin is not present
        with pytest.raises(PluginNotFoundError):
            project.plugins.find_plugin(plugin_name, plugin_type=plugin_type)

        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(cli, ["add", plugin_type.singular, plugin_name])

            if plugin_type is PluginType.TRANSFORMS:
                assert res.exit_code == 1, res.stdout
                assert isinstance(res.exception, CliError)
                assert "Dependencies not met:" in str(res.exception)
            else:
                assert res.exit_code == 0, res.stdout
                assert f"Added {plugin_type.descriptor} '{plugin_name}'" in res.stdout

                plugin = project.plugins.find_plugin(plugin_name, plugin_type)
                assert plugin
                assert plugin.variant == default_variant

                plugins = [plugin]

                for required_plugin_ref in required_plugin_refs:
                    if (required_plugin_ref._type) == PluginType.FILES and (
                        required_plugin_ref.name == "dbt"
                    ):
                        # File bundles with no managed files are added but
                        # do not appear in meltano.yml
                        assert (
                            f"Adding required file bundle '{required_plugin_ref.name}'"
                            in res.stdout
                        )
                    else:
                        plugin = project.plugins.get_plugin(required_plugin_ref)
                        assert plugin

                        assert (
                            f"Added required {plugin.type.descriptor} '{plugin.name}'"
                            in res.stdout
                        )

                        plugins.append(plugin)

                install_plugin_mock.assert_called()

    @pytest.mark.order(1)
    def test_add_multiple(self, project: Project, cli_runner):
        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            cli_runner.invoke(cli, ["add", "extractors", "tap-gitlab"])

        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            res = cli_runner.invoke(
                cli,
                ["add", "extractors", "tap-gitlab", "tap-adwords", "tap-facebook"],
            )

            assert res.exit_code == 0, res.stdout
            assert (
                "Extractor 'tap-gitlab' already exists in your Meltano project"
                in res.stderr
            )
            assert "Added extractor 'tap-adwords'" in res.stdout
            assert "Added extractor 'tap-facebook'" in res.stdout

            tap_gitlab = project.plugins.find_plugin(
                "tap-gitlab",
                PluginType.EXTRACTORS,
            )
            assert tap_gitlab
            tap_adwords = project.plugins.find_plugin(
                "tap-adwords",
                PluginType.EXTRACTORS,
            )
            assert tap_adwords
            tap_facebook = project.plugins.find_plugin(
                "tap-facebook",
                PluginType.EXTRACTORS,
            )
            assert tap_facebook

            install_plugin_mock.assert_called_once_with(
                project,
                [tap_gitlab, tap_adwords, tap_facebook],
                reason=PluginInstallReason.ADD,
                force=False,
            )

    @pytest.mark.order(1)
    def test_add_different_variant(self, cli_runner):
        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(cli, ["add", "extractor", "tap-mock"])
            assert res.exit_code == 0, res.stdout
            assert "Added extractor 'tap-mock" in res.stdout

            res = cli_runner.invoke(
                cli,
                ["add", "extractor", "tap-mock", "--variant", "singer-io"],
            )
            assert res.exit_code == 0, res.stdout
            assert (
                "Extractor 'tap-mock' already exists in your Meltano project"
                in res.stderr
            )
            assert "To switch from the current" in res.stdout
            assert "name: tap-mock" in res.stdout
            assert "variant: singer-io" in res.stdout

    @pytest.mark.order(2)
    def test_add_transform(self, project: Project, cli_runner):
        # adding Transforms requires the legacy 'dbt' Transformer
        cli_runner.invoke(cli, ["add", "transformer", "dbt"])
        cli_runner.invoke(cli, ["install", "transformer", "dbt"])

        res = cli_runner.invoke(cli, ["add", "transform", "tap-google-analytics"])
        assert_cli_runner(res)

        with project.root_dir("transform/packages.yml").open() as packages_file:
            packages_yaml = yaml.safe_load(packages_file)

        with project.root_dir("transform/dbt_project.yml").open() as project_file:
            project_yaml = yaml.safe_load(project_file)

        assert {
            "git": "https://gitlab.com/meltano/dbt-tap-google-analytics.git",
            "revision": "config-version-2",
        } in packages_yaml["packages"]

        assert "tap_google_analytics" in project_yaml["models"]
        assert project_yaml["vars"]["tap_google_analytics"] == {
            "schema": "{{ env_var('DBT_SOURCE_SCHEMA', 'tap_google_analytics') }}",
        }

    @fails_on_windows
    def test_add_files_with_updates(
        self,
        project: Project,
        cli_runner,
        plugin_settings_service_factory,
    ):
        # if plugin is locked, we actually wouldn't expect it to update.
        # So we must remove lockfile
        shutil.rmtree("plugins/files", ignore_errors=True)

        result = cli_runner.invoke(cli, ["add", "files", "airflow"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        # Plugin has been added to meltano.yml
        plugin = project.plugins.find_plugin("airflow", PluginType.FILES)
        assert plugin

        # Automatic updating is enabled
        plugin_settings_service = plugin_settings_service_factory(plugin)
        update_config = plugin_settings_service.get("_update")
        assert update_config["orchestrate/dags/meltano.py"] is True

        # File has been created
        assert "Created orchestrate/dags/meltano.py" in output

        file_path = project.root_dir("orchestrate/dags/meltano.py")
        assert file_path.is_file()

        # File has "managed" header
        assert (
            "This file is managed by the 'airflow' file bundle" in file_path.read_text()
        )

    def test_add_files_without_updates(self, project: Project, cli_runner):
        result = cli_runner.invoke(cli, ["add", "files", "docker-compose"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        # Plugin has not been added to meltano.yml
        with pytest.raises(PluginNotFoundError):
            project.plugins.find_plugin("docker-compose", PluginType.FILES)

        # File has been created
        assert "Created docker-compose.yml" in output

        file_path = project.root_dir("docker-compose.yml")
        assert file_path.is_file()

        # File does not have "managed" header
        assert "This file is managed" not in file_path.read_text()

    @fails_on_windows
    def test_add_files_that_already_exists(self, project: Project, cli_runner):
        # dbt lockfile was created in an upstream test. Need to remove.
        shutil.rmtree(project.root_dir("plugins/files"), ignore_errors=True)
        project.root_dir("transform/dbt_project.yml").write_text("Exists!")
        result = cli_runner.invoke(cli, ["add", "files", "dbt"])
        output = result.stdout + result.stderr
        assert_cli_runner(result)

        assert (
            "File 'transform/dbt_project.yml' already exists, keeping both versions"
            in output
        )
        assert "Created transform/dbt_project (dbt).yml" in output
        assert project.root_dir("transform/dbt_project (dbt).yml").is_file()

    def test_add_missing(self, project: Project, cli_runner):
        res = cli_runner.invoke(cli, ["add", "extractor", "tap-unknown"])

        assert res.exit_code == 1
        assert res.exception
        assert str(res.exception) == (
            "Extractor 'tap-unknown' is not known to Meltano. "
            "Check https://hub.meltano.com/ for available plugins."
        )

        # ensure the plugin is not present
        with pytest.raises(PluginNotFoundError):
            project.plugins.find_plugin("tap-unknown", PluginType.EXTRACTORS)

    @pytest.mark.xfail(reason="Uninstall not implemented yet.")
    def test_add_fails(self, project: Project, cli_runner):
        result = cli_runner.invoke(cli, ["add", "extractor", "tap-mock"])
        output = result.stdout + result.stderr

        assert result.exit_code == 1, result.stdout
        assert "Failed to install plugin 'tap-mock'" in output
        assert result.stderr

        # ensure the plugin is not present
        with pytest.raises(PluginNotFoundError):
            project.plugins.find_plugin("tap-mock", PluginType.EXTRACTORS)

    def test_add_variant(self, project: Project, cli_runner):
        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(
                cli,
                [
                    "add",
                    "mapper",
                    "mapper-mock",
                    "--variant",
                    "alternative",
                ],
            )
            assert_cli_runner(res)

            plugin = project.plugins.find_plugin(
                plugin_type=PluginType.MAPPERS,
                plugin_name="mapper-mock",
            )
            assert plugin.variant == "alternative"

    def test_add_inherited(
        self,
        project: Project,
        tap,
        cli_runner,
    ):
        # Make sure tap-mock is not in the project as a project plugin
        project.plugins.remove_from_file(tap)

        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True

            # Inheriting from a BasePlugin using --as
            res = cli_runner.invoke(
                cli,
                [
                    "add",
                    "extractor",
                    "tap-mock",
                    "--as",
                    "tap-mock-inherited",
                ],
            )
            assert_cli_runner(res)
            assert "Inherit from:\ttap-mock, variant meltano (default)\n" in res.stdout

            inherited = project.plugins.find_plugin(
                plugin_type=PluginType.EXTRACTORS,
                plugin_name="tap-mock-inherited",
            )
            assert inherited.inherit_from == "tap-mock"
            assert inherited.variant == "meltano"
            assert inherited.parent == project.hub_service.find_base_plugin(
                plugin_type=PluginType.EXTRACTORS,
                plugin_name="tap-mock",
                variant="meltano",
            )

            # Inheriting from a BasePlugin using --inherit-from and --variant
            res = cli_runner.invoke(
                cli,
                [
                    "add",
                    "extractor",
                    "tap-mock--singer-io",
                    "--inherit-from",
                    "tap-mock",
                    "--variant",
                    "singer-io",
                ],
            )
            assert_cli_runner(res)
            assert (
                "Inherit from:\ttap-mock, variant singer-io (deprecated)\n"
                in res.stdout
            )

            inherited_variant = project.plugins.find_plugin(
                plugin_type=PluginType.EXTRACTORS,
                plugin_name="tap-mock--singer-io",
            )
            assert inherited_variant.inherit_from == "tap-mock"
            assert inherited_variant.variant == "singer-io"
            assert inherited_variant.parent == project.hub_service.find_base_plugin(
                plugin_type=PluginType.EXTRACTORS,
                plugin_name="tap-mock",
                variant="singer-io",
            )

            # Inheriting from a ProjectPlugin using --inherit-from
            res = cli_runner.invoke(
                cli,
                [
                    "add",
                    "extractor",
                    "tap-mock-inception",
                    "--inherit-from",
                    "tap-mock-inherited",
                ],
            )
            assert_cli_runner(res)
            assert "Inherit from:\ttap-mock-inherited\n" in res.stdout

            inception = project.plugins.find_plugin(
                plugin_type=PluginType.EXTRACTORS,
                plugin_name="tap-mock-inception",
            )
            assert inception.inherit_from == "tap-mock-inherited"
            assert inception.parent == inherited

            # Inheriting from a nonexistent plugin
            res = cli_runner.invoke(
                cli,
                [
                    "add",
                    "extractor",
                    "tap-foo",
                    "--inherit-from",
                    "tap-bar",
                ],
            )
            assert res.exit_code == 1
            assert (
                "Could not find parent plugin for extractor 'tap-foo': "
                "Extractor 'tap-bar' is not known to Meltano"
            ) in str(res.exception)

    @pytest.mark.usefixtures("reset_project_context")
    def test_add_custom(self, project: Project, cli_runner):
        pip_url = "-e path/to/tap-custom"
        executable = "tap-custom-bin"
        stdin = os.linesep.join(
            # namespace, pip_url, executable, capabilities, settings
            ["tap_custom", pip_url, executable, "foo,bar", "baz,qux"],
        )

        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(
                cli,
                ["add", "--custom", "extractor", "tap-custom"],
                input=stdin,
            )
            assert_cli_runner(res)

            plugin = project.plugins.find_plugin(
                plugin_type=PluginType.EXTRACTORS,
                plugin_name="tap-custom",
            )
            assert plugin.name == "tap-custom"
            assert plugin.pip_url == pip_url

            plugin_def = plugin.custom_definition
            plugin_variant = plugin.custom_definition.variants[0]

            assert plugin_def.type == plugin.type
            assert plugin_def.name == plugin.name == "tap-custom"
            assert plugin_def.namespace == plugin.namespace == "tap_custom"

            assert plugin_variant.name is None

            assert plugin.pip_url == plugin_variant.pip_url == pip_url
            assert plugin.executable == plugin_variant.executable == executable
            assert plugin.capabilities == plugin_variant.capabilities == ["foo", "bar"]

            assert [stg.name for stg in plugin_variant.settings] == ["baz", "qux"]
            assert plugin.settings == plugin_variant.settings

            install_plugin_mock.assert_called_once_with(
                project,
                [plugin],
                reason=PluginInstallReason.ADD,
                force=False,
            )

    def test_add_custom_no_install(self, project: Project, cli_runner):
        executable = "tap-custom-noinstall"
        stdin = os.linesep.join(
            # namespace, pip_url, executable, capabilities, settings
            [
                "tap_custom_noinstall",
                "n",
                executable,
                "foo,bar",
                "baz,qux",
            ],
        )

        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(
                cli,
                ["add", "--custom", "extractor", executable],
                input=stdin,
            )
            assert_cli_runner(res)

            plugin: ProjectPlugin = project.plugins.find_plugin(
                plugin_type=PluginType.EXTRACTORS,
                plugin_name=executable,
            )
            assert plugin.name == executable
            assert plugin.is_installable() is False
            assert plugin.is_invokable() is True
            assert plugin.pip_url is None

            plugin_def = plugin.custom_definition
            plugin_variant = plugin.custom_definition.variants[0]

            assert plugin_def.type == plugin.type
            assert plugin_def.name == plugin.name == executable

            assert plugin_variant.name is None

            assert plugin.pip_url is None
            assert plugin_variant.pip_url is None
            assert plugin.executable == plugin_variant.executable == executable

            install_plugin_mock.assert_called_once_with(
                project,
                [plugin],
                reason=PluginInstallReason.ADD,
                force=False,
            )

    def test_add_custom_variant(self, project: Project, cli_runner):
        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(
                cli,
                [
                    "add",
                    "--custom",
                    "extractor",
                    "tap-custom-variant",
                    "--variant",
                    "personal",
                ],
            )
            assert_cli_runner(res)

            plugin = project.plugins.find_plugin(
                plugin_type=PluginType.EXTRACTORS,
                plugin_name="tap-custom-variant",
            )
            plugin_def = plugin.custom_definition
            plugin_variant = plugin_def.variants[0]

            assert plugin.variant == plugin_variant.name == "personal"

    @pytest.mark.parametrize(
        ("plugin_type", "plugin_name", "default_variant", "required_plugin_refs"),
        (
            (PluginType.EXTRACTORS, "tap-carbon-intensity", "meltano", []),
            (PluginType.LOADERS, "target-sqlite", "meltanolabs", []),
            (PluginType.TRANSFORMS, "tap-carbon-intensity", "meltano", []),
            (
                PluginType.ORCHESTRATORS,
                "airflow",
                Variant.ORIGINAL_NAME,
                [PluginRef(PluginType.FILES, "airflow")],
            ),
        ),
        ids=[
            "single-extractor",
            "single-loader",
            "transform-and-related",
            "orchestrator-and-required",
        ],
    )
    @pytest.mark.usefixtures("reset_project_context")
    def test_add_no_install(
        self,
        plugin_type,
        plugin_name,
        default_variant,
        required_plugin_refs,
        project: Project,
        cli_runner,
    ):
        # ensure the plugin is not present
        with pytest.raises(PluginNotFoundError):
            project.plugins.find_plugin(plugin_name, plugin_type=plugin_type)

        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(
                cli,
                ["add", plugin_type.singular, plugin_name, "--no-install"],
            )

            if plugin_type is PluginType.TRANSFORMS:
                assert res.exit_code == 1, res.stdout
                assert isinstance(res.exception, CliError)
                assert "Dependencies not met:" in str(res.exception)
            else:
                assert res.exit_code == 0, res.stdout
                assert f"Added {plugin_type.descriptor} '{plugin_name}'" in res.stdout

                plugin = project.plugins.find_plugin(plugin_name, plugin_type)
                assert plugin
                assert plugin.variant == default_variant

                # check plugin lock file is added
                plugins_dir = project.root_dir("plugins")
                assert plugins_dir.joinpath(
                    f"{plugin_type}/{plugin_name}--{default_variant}.lock",
                ).exists()

                for required_plugin_ref in required_plugin_refs:
                    if (required_plugin_ref._type) == PluginType.FILES and (
                        required_plugin_ref.name == "dbt"
                    ):
                        # File bundles with no managed files are added but do
                        # not appear in meltano.yml
                        assert (
                            f"Adding required file bundle '{required_plugin_ref.name}'"
                            in res.stdout
                        )
                    else:
                        plugin = project.plugins.get_plugin(required_plugin_ref)
                        assert plugin

                        assert (
                            f"Added required {plugin.type.descriptor} '{plugin.name}'"
                            in res.stdout
                        )

                    # check required plugin lock files are added
                    assert list(
                        plugins_dir.glob(
                            f"{required_plugin_ref._type}/"
                            f"{required_plugin_ref.name}--*.lock",
                        ),
                    )

                install_plugin_mock.assert_not_called()

    @pytest.mark.parametrize(
        "ref",
        (
            plugin_ref,
            (
                "https://raw.githubusercontent.com/meltano/hub/main/_data/meltano/"
                f"{plugin_ref.relative_to(plugins_dir)}"
            ),
        ),
        ids=(
            "local",
            "remote",
        ),
    )
    @pytest.mark.usefixtures("reset_project_context")
    @mock.patch("meltano.cli.add.install_plugins")
    @mock.patch("meltano.cli.add.requests.get")
    def test_add_from_ref(
        self,
        ref_request_mock,
        install_plugin_mock,
        ref,
        project,
        cli_runner,
    ):
        ref_request_mock.return_value.status_code = 200
        ref_request_mock.return_value.text = plugin_ref.read_text()

        install_plugin_mock.return_value = True

        plugin_name = plugin_ref.parent.name
        plugin_type = PluginType(plugin_ref.parents[1].name)

        res = cli_runner.invoke(
            cli,
            ["add", plugin_type.singular, plugin_name, "--from-ref", ref],
        )
        assert_cli_runner(res)

        assert f"Added {plugin_type.singular} '{plugin_name}'" in res.output

        plugin = project.plugins.find_plugin(plugin_name, plugin_type)
        assert plugin.name == plugin_name
        assert plugin.variant == "test"

    @pytest.mark.usefixtures("reset_project_context")
    @pytest.mark.parametrize(
        (
            "ref",
            "invalid_reason",
        ),
        (
            (
                plugin_ref.name,
                "No such file or directory: '{ref}'",
            ),
            pytest.param(
                plugin_ref.parent,
                "Is a directory: '{ref}'",
                marks=fails_on_windows,
            ),
            (
                "https://:",
                "Invalid URL '{ref}'",
            ),
        ),
        ids=(
            "does not exist",
            "is directory",
            "invalid url",
        ),
    )
    def test_add_from_ref_invalid_ref(
        self,
        ref,
        invalid_reason,
        cli_runner,
    ):
        res = cli_runner.invoke(
            cli,
            ["add", "extractor", "tap-custom", "--from-ref", ref],
        )

        assert res.exit_code == 2
        assert invalid_reason.format(ref=ref) in res.stderr

    @pytest.mark.parametrize(
        (
            "definition",
            "invalid_reason",
        ),
        (
            (
                "test",
                "incorrect format",
            ),
            (
                {},
                "missing properties (name, namespace)",
            ),
            (
                {"test-key": "test-value"},
                "missing properties (name, namespace)",
            ),
            (
                {"name": "tap-custom"},
                "missing properties (namespace)",
            ),
        ),
        ids=(
            "incorrect format",
            "empty",
            "no required properties",
            "some required properties",
        ),
    )
    def test_add_from_ref_invalid_definiton(
        self,
        definition,
        invalid_reason,
        cli_runner,
    ):
        with open("test.yml", "w") as f:
            yaml.dump(definition, f)

        res = cli_runner.invoke(
            cli,
            ["add", "extractor", "tap-custom", "--from-ref", f.name],
        )

        assert res.exit_code == 1
        assert isinstance(res.exception, InvalidPluginDefinitionError)
        assert res.exception.reason == invalid_reason

    def test_add_with_python_version(self, cli_runner: CliRunner):
        with mock.patch(
            "meltano.core.venv_service.VirtualEnv._resolve_python_path",
        ) as venv_mock, mock.patch("meltano.core.venv_service.VenvService.install"):
            python = "python3.X"
            assert_cli_runner(
                cli_runner.invoke(
                    cli,
                    (
                        "add",
                        "extractor",
                        "tap-that-needs-custom-python",
                        "--python",
                        python,
                    ),
                ),
            )
            venv_mock.assert_called_once_with(python)

    def test_add_with_force_flag(self, project: Project, cli_runner: CliRunner):
        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(
                cli,
                ["add", "extractor", "tap-gitlab", "--force-install"],
            )
            tap_gitlab = project.plugins.find_plugin(
                "tap-gitlab",
                PluginType.EXTRACTORS,
            )

        assert_cli_runner(res)
        assert tap_gitlab
        install_plugin_mock.assert_called_once_with(
            project,
            [tap_gitlab],
            reason=PluginInstallReason.ADD,
            force=True,
        )

    @pytest.mark.usefixtures("reset_project_context")
    def test_add_update(self, cli_runner):
        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(cli, ["add", "extractor", "tap-mock"])
            assert res.exit_code == 0, res.stdout
            assert "Added extractor 'tap-mock" in res.stdout

            res = cli_runner.invoke(
                cli,
                ["add", "extractor", "tap-mock", "--update"],
            )
            assert res.exit_code == 0, res.stdout
            assert "Updated extractor 'tap-mock" in res.stdout

    @pytest.mark.usefixtures("reset_project_context")
    def test_add_update_not_in_project(self, cli_runner):
        res = cli_runner.invoke(cli, ["add", "extractor", "tap-mock", "--update"])

        assert res.exit_code == 1
        assert res.exception
        assert str(res.exception) == "Extractor 'tap-mock' is not known to Meltano"

    def test_lockfile_exists(self, cli_runner):
        plugins_dir = Path("plugins/utilities")
        plugins_dir.mkdir(parents=True, exist_ok=True)
        lockfile = plugins_dir / "utility-mock--original.lock"
        lockfile.write_text(
            json.dumps(
                {
                    "plugin_type": "utilities",
                    "name": "utility-mock",
                    "namespace": "utility_mock",
                }
            )
        )
        lockfile.touch()

        with mock.patch("meltano.cli.add.install_plugins") as install_plugin_mock:
            install_plugin_mock.return_value = True
            res = cli_runner.invoke(cli, ["add", "utility", "utility-mock"])
            assert res.exit_code == 0, res.stdout
            assert "Plugin definition is already locked at" in res.stdout
            assert "You can remove the file manually" in res.stdout



"""Schedule management CLI."""

from __future__ import annotations

import json
import sys
import typing as t

import click
from croniter import croniter

from meltano.cli.params import pass_project
from meltano.cli.utils import (
    CliEnvironmentBehavior,
    InstrumentedDefaultGroup,
    PartialInstrumentedCmd,
)
from meltano.core.db import project_engine
from meltano.core.job.stale_job_failer import fail_stale_jobs
from meltano.core.schedule import CRON_INTERVALS
from meltano.core.schedule_service import (
    BadCronError,
    ScheduleAlreadyExistsError,
    ScheduleService,
)
from meltano.core.task_sets_service import TaskSetsService
from meltano.core.utils import coerce_datetime

if t.TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from meltano.core.project import Project
    from meltano.core.schedule import Schedule
    from meltano.core.task_sets import TaskSets


@click.group(
    cls=InstrumentedDefaultGroup,
    default="add",
    short_help="Manage pipeline schedules.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_ignore_default,
)
@click.pass_context
@pass_project(migrate=True)
def schedule(project, ctx):
    """
    Manage pipeline schedules.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#schedule
    """
    ctx.obj["project"] = project
    ctx.obj["schedule_service"] = ScheduleService(project)
    ctx.obj["task_sets_service"] = TaskSetsService(project)


def _add_elt(
    ctx,
    name: str,
    extractor: str,
    loader: str,
    transform: str,
    interval: str,
    start_date: str | None,
):
    """Add a new legacy elt schedule."""
    project: Project = ctx.obj["project"]
    schedule_service: ScheduleService = ctx.obj["schedule_service"]

    _, session_maker = project_engine(project)
    session = session_maker()
    try:
        added_schedule = schedule_service.add_elt(
            session,
            name,
            extractor,
            loader,
            transform,
            interval,
            start_date,
        )
        click.echo(
            f"Scheduled elt '{added_schedule.name}' at {added_schedule.interval}",
        )
    except ScheduleAlreadyExistsError:
        click.secho(f"Schedule '{name}' already exists.", fg="yellow")
    finally:
        session.close()


def _add_job(ctx, name: str, job: str, interval: str):
    """Add a new scheduled job."""
    project: Project = ctx.obj["project"]
    schedule_service: ScheduleService = ctx.obj["schedule_service"]

    _, session_maker = project_engine(project)
    session = session_maker()
    try:
        added_schedule = schedule_service.add(name, job, interval)
        click.echo(
            f"Scheduled job '{added_schedule.name}' at {added_schedule.interval}",
        )
    except ScheduleAlreadyExistsError:
        click.secho(f"Schedule '{name}' already exists.", fg="yellow")
    finally:
        session.close()


@schedule.command(
    cls=PartialInstrumentedCmd,
    short_help="[default] Add a new schedule.",
)
@click.argument("name")
@click.option("--interval", required=True, help="Interval of the schedule.")
@click.option("--job", help="The name of the job to run.")
@click.option("--extractor", required=False, help="ELT Only")
@click.option("--loader", required=False, help="ELT Only")
@click.option(
    "--transform",
    type=click.Choice(["skip", "only", "run"]),
    default="skip",
    help="ELT Only",
)
@click.option("--start-date", type=click.DateTime(), default=None, help="ELT Only")
@click.pass_context
def add(ctx, name, job, extractor, loader, transform, interval, start_date):
    """
    Add a new schedule. Schedules can be used to run Meltano jobs or ELT tasks at a specific interval.

    Example usage:

    \b
    \t# Schedule a job name "my_job" to run everyday
    \tmeltano schedule add <schedule_name> --job my_job --interval "@daily"
    \t# Schedule an ELT task to run hourly
    \tmeltano schedule add <schedule_name> --extractor <tap> --loader <target> --transform run --interval "@hourly"

    \b\nNote that the --job option and --extractor/--loader options are mutually exclusive.

    \b\nRead more at https://docs.meltano.com/reference/command-line-interface#schedule
    """  # noqa: E501
    if job and (extractor or loader):
        raise click.ClickException(
            "Cannot mix --job with --extractor/--loader/--transform",  # noqa: EM101
        )

    if not job:
        if not extractor:
            raise click.ClickException("Missing --extractor")  # noqa: EM101
        if not loader:
            raise click.ClickException("Missing --loader")  # noqa: EM101

        _add_elt(ctx, name, extractor, loader, transform, interval, start_date)
        return
    _add_job(ctx, name, job, interval)


def _format_job_list_output(entry: Schedule, job: TaskSets) -> dict:
    return {
        "name": entry.name,
        "interval": entry.interval,
        "cron_interval": entry.cron_interval,
        "env": entry.env,
        "job": {
            "name": job.name,
            "tasks": job.tasks,
        },
    }


def _format_elt_list_output(entry: Schedule, session: Session) -> dict:
    start_date = coerce_datetime(entry.start_date)
    if start_date:
        start_date = start_date.date().isoformat()

    last_successful_run = entry.last_successful_run(session)
    last_successful_run_ended_at = (
        last_successful_run.ended_at.isoformat() if last_successful_run else None
    )

    return {
        "name": entry.name,
        "extractor": entry.extractor,
        "loader": entry.loader,
        "transform": entry.transform,
        "interval": entry.interval,
        "start_date": start_date,
        "env": entry.env,
        "cron_interval": entry.cron_interval,
        "last_successful_run_ended_at": last_successful_run_ended_at,
        "elt_args": entry.elt_args,
    }


@schedule.command(  # noqa: WPS125
    cls=PartialInstrumentedCmd,
    short_help="List available schedules.",
)
@click.option("--format", type=click.Choice(("json", "text")), default="text")
@click.pass_context
def list(ctx, format):  # noqa: WPS125
    """List available schedules."""
    project = ctx.obj["project"]
    schedule_service: ScheduleService = ctx.obj["schedule_service"]
    task_sets_service: TaskSetsService = ctx.obj["task_sets_service"]

    _, sessionMaker = project_engine(project)  # noqa: N806
    session = sessionMaker()
    try:
        fail_stale_jobs(session)

        if format == "text":
            transform_elt_markers = {
                "run": ("→", "→"),
                "only": ("×", "→"),
                "skip": ("→", "x"),
            }

            for txt_schedule in schedule_service.schedules():
                if txt_schedule.job:
                    click.echo(
                        f"[{txt_schedule.interval}] job {txt_schedule.name}: "
                        f"{txt_schedule.job} → "
                        f"{task_sets_service.get(txt_schedule.job).tasks}",
                    )
                else:
                    markers = transform_elt_markers[txt_schedule.transform]
                    click.echo(
                        f"[{txt_schedule.interval}] elt {txt_schedule.name}: "
                        f"{txt_schedule.extractor} {markers[0]} "
                        f"{txt_schedule.loader} {markers[1]} transforms",
                    )

        elif format == "json":
            job_schedules = []
            elt_schedules = []
            for json_schedule in schedule_service.schedules():
                if json_schedule.job:
                    job_schedules.append(
                        _format_job_list_output(
                            json_schedule,
                            task_sets_service.get(json_schedule.job),
                        ),
                    )
                else:
                    elt_schedules.append(
                        _format_elt_list_output(json_schedule, session),
                    )
            click.echo(
                json.dumps(
                    {"schedules": {"job": job_schedules, "elt": elt_schedules}},
                    indent=2,
                ),
            )
    finally:
        session.close()


@schedule.command(
    cls=PartialInstrumentedCmd,
    context_settings={"ignore_unknown_options": True, "allow_interspersed_args": False},
    short_help="Run a schedule.",
)
@click.argument("name")
@click.argument("elt_options", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def run(ctx, name, elt_options):
    """Run a schedule."""
    schedule_service: ScheduleService = ctx.obj["schedule_service"]
    process = schedule_service.run(schedule_service.find_schedule(name), *elt_options)
    if process.returncode:
        sys.exit(process.returncode)


@schedule.command(
    cls=PartialInstrumentedCmd,
    name="remove",
    short_help="Remove a schedule.",
)
@click.argument("name", required=True)
@click.pass_context
def remove(ctx, name):
    """Remove a schedule.

    Usage:
        meltano schedule remove <name>
    """
    ctx.obj["schedule_service"].remove(name)


def _update_job_schedule(
    candidate: Schedule,
    job: str | None,
    interval: str | None = None,
) -> Schedule:
    """Update an existing job schedule.

    Args:
        candidate: The schedule to update.
        job: The name of the job to run.
        interval: The interval of the schedule.

    Raises:
        click.ClickException: If the schedule is not a scheduled job.

    Returns:
        The updated schedule.
    """
    if not candidate.job:
        raise click.ClickException(
            f"Cannot update schedule {candidate.name} with job only flags as "  # noqa: EM102
            "its a elt schedule",
        )
    if job:
        candidate.job = job
    if interval:
        candidate.interval = interval
    return candidate


def _update_elt_schedule(
    candidate: Schedule,
    extractor: str | None,
    loader: str | None,
    transform: str | None,
    interval: str | None,
) -> Schedule:
    """Update an elt schedule.

    Args:
        candidate: The schedule to update.
        extractor: The name of the extractor to use.
        loader: The name of the loader to use.
        transform: The transform flag to use.
        interval: The interval of the schedule.

    Raises:
        click.ClickException: If the schedule is not a scheduled elt task.

    Returns:
        The updated schedule.
    """
    if candidate.job:
        raise click.ClickException(
            f"Cannot update schedule {candidate.name} with elt only flags as "  # noqa: EM102
            "its a scheduled job",
        )

    if extractor:
        candidate.extractor = extractor
    if loader:
        candidate.loader = loader
    if transform:
        candidate.transform = transform
    if interval:
        candidate.interval = interval
    return candidate


class CronParam(click.ParamType):
    """Custom type definition for cron prameter."""

    name = "cron"

    def convert(self, value, *_):
        """Validate and con interval."""
        if value not in CRON_INTERVALS and not croniter.is_valid(value):
            raise BadCronError(value)

        return value


@schedule.command(
    cls=PartialInstrumentedCmd,
    name="set",
    short_help="Update a schedule.",
)
@click.argument("name", required=True)
@click.option(
    "--interval",
    help="Update the interval of the schedule.",
    type=CronParam(),
)
@click.option("--job", help="Update the name of the job to run a scheduled job.")
@click.option("--extractor", help="Update the extractor for an elt schedule.")
@click.option("--loader", help="Updated the loader for an elt schedule.")
@click.option(
    "--transform",
    type=click.Choice(["skip", "only", "run"]),
    default=None,
    help="Update the transform flag for an elt schedule.",
)
@click.pass_context
def set_cmd(ctx, name, interval, job, extractor, loader, transform):
    """Update a schedule.

    Usage:
        meltano schedule set <name> [--interval <interval>] [--job <job>] [--extractor <extractor>] [--loader <loader>] [--transform <transform>]
    """  # noqa: E501
    schedule_service: ScheduleService = ctx.obj["schedule_service"]
    candidate = schedule_service.find_schedule(name)

    if candidate.job:
        if extractor or loader or transform:
            raise click.ClickException(
                "Cannot mix --job with --extractor/--loader/--transform",  # noqa: EM101
            )
        updated = _update_job_schedule(candidate, job, interval)
    else:
        if job:
            raise click.ClickException(
                "Cannot mix --job with --extractor/--loader/--transform",  # noqa: EM101
            )
        updated = _update_elt_schedule(
            candidate,
            extractor,
            loader,
            transform,
            interval,
        )

    schedule_service.update_schedule(updated)
    click.echo(f"Updated schedule '{name}'")



from __future__ import annotations

import pytest

from meltano.core.meltano_file import MeltanoFile


class TestMeltanoFile:
    @pytest.mark.usefixtures("tap", "target")
    def test_load(self, project):
        meltano_file = MeltanoFile.parse(project.meltano)
        assert meltano_file

    @pytest.mark.usefixtures("mapper")
    def test_get_plugins_for_mappings(self, project):
        meltano_file = MeltanoFile.parse(project.meltano)

        test_config = {
            "name": "mapper-mock",
            "variant": "meltano",
            "mappings": [
                {
                    "name": "mock-mapping-0",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "author_email",
                                "tap_stream_name": "commits",
                                "type": "MASK-HIDDEN",
                            },
                        ],
                    },
                },
                {
                    "name": "mock-mapping-1",
                    "config": {
                        "transformations": [
                            {
                                "field_id": "given_name",
                                "tap_stream_name": "users",
                                "type": "lowercase",
                            },
                        ],
                    },
                },
            ],
        }

        plugins = meltano_file.get_plugins_for_mappings(test_config)
        assert len(plugins) == 2

        assert plugins[0].extra_config.get("_mapping")
        assert (
            plugins[0].extra_config.get("_mapping_name")
            == test_config["mappings"][0]["name"]
        )
        assert plugins[0].config == test_config["mappings"][0]["config"]

        assert plugins[1].extra_config.get("_mapping")
        assert (
            plugins[1].extra_config.get("_mapping_name")
            == test_config["mappings"][1]["name"]
        )
        assert plugins[1].config == test_config["mappings"][1]["config"]



from __future__ import annotations

import platform

import dotenv
import pytest

from meltano.core.plugin.command import UndefinedEnvVarError
from meltano.core.plugin_invoker import UnknownCommandError
from meltano.core.tracking.contexts import environment_context
from meltano.core.venv_service import VirtualEnv


class TestPluginInvoker:
    @pytest.fixture()
    async def plugin_invoker(self, utility, session, plugin_invoker_factory):
        subject = plugin_invoker_factory(utility)
        async with subject.prepared(session):
            yield subject

    @pytest.fixture()
    async def nonpip_plugin_invoker(self, nonpip_tap, session, plugin_invoker_factory):
        subject = plugin_invoker_factory(nonpip_tap)
        async with subject.prepared(session):
            yield subject

    @pytest.mark.asyncio()
    async def test_env(self, project, tap, session, plugin_invoker_factory):
        project.dotenv.touch()
        dotenv.set_key(project.dotenv, "DUMMY_ENV_VAR", "from_dotenv")
        dotenv.set_key(project.dotenv, "TAP_MOCK_TEST", "from_dotenv")
        project.refresh()

        subject = plugin_invoker_factory(tap)
        async with subject.prepared(session):
            env = subject.env()

        # .env
        assert env["DUMMY_ENV_VAR"] == "from_dotenv"

        # Project env
        assert env["MELTANO_PROJECT_ROOT"] == str(project.root)
        assert env["MELTANO_ENVIRONMENT"] == ""

        # Project settings
        assert env["MELTANO_CLI_LOG_LEVEL"] == "info"

        # Plugin info
        assert env["MELTANO_EXTRACTOR_NAME"] == tap.name

        # Plugin settings
        assert env["MELTANO_EXTRACT_TEST"] == env["TAP_MOCK_TEST"] == "from_dotenv"
        assert env["MELTANO_EXTRACT__SELECT"] == env["TAP_MOCK__SELECT"] == '["*.*"]'

        # Plugin execution environment
        venv = VirtualEnv(project.venvs_dir(tap.type, tap.name))
        assert env["VIRTUAL_ENV"] == str(venv.root)
        assert env["PATH"].startswith(str(venv.bin_dir))
        assert "PYTHONPATH" not in env

        assert (
            env["MELTANO_PARENT_CONTEXT_UUID"]
            == environment_context.data["context_uuid"]
        )

    @pytest.mark.asyncio()
    async def test_environment_env(
        self,
        project_with_environment,
        tap,
        session,
        plugin_invoker_factory,
    ):
        subject = plugin_invoker_factory(tap)
        async with subject.prepared(session):
            env = subject.env()

        # Project env
        assert env["MELTANO_ENVIRONMENT"] == project_with_environment.environment.name

    @pytest.mark.asyncio()
    async def test_expanded_environment_env(
        self,
        project_with_environment,
        tap,
        session,
        plugin_invoker_factory,
    ):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        subject = plugin_invoker_factory(tap)
        async with subject.prepared(session):
            env = subject.env()

        assert env["ENVIRONMENT_ENV_VAR"] == str(
            project_with_environment.root / "file.txt",
        )

    @pytest.mark.asyncio()
    async def test_unknown_command(self, plugin_invoker):
        with pytest.raises(UnknownCommandError) as err:
            await plugin_invoker.invoke_async(command="foo")

        assert err.value.command == "foo"
        assert "supports the following commands" in str(err.value)

    def test_expand_exec_args(self, plugin_invoker):
        exec_args = plugin_invoker.exec_args(
            "extra",
            "args",
            command="cmd",
            env={
                "ENV_VAR_ARG": "env-var-arg",
            },
        )

        assert exec_args[0].endswith("utility-mock")
        assert exec_args[1:] == ["--option", "env-var-arg", "extra", "args"]

    def test_expand_command_exec_args(self, plugin_invoker):
        exec_args = plugin_invoker.exec_args(
            "extra",
            "args",
            command="cmd",
            env={
                "ENV_VAR_ARG": "env-var-arg",
            },
        )

        assert exec_args[0].endswith("utility-mock")
        assert exec_args[1:] == ["--option", "env-var-arg", "extra", "args"]

    @pytest.mark.asyncio()
    async def test_undefined_env_var(self, plugin_invoker):
        with pytest.raises(UndefinedEnvVarError) as err:
            await plugin_invoker.invoke_async(command="cmd")

        assert (
            "Command 'cmd' referenced unset environment "
            "variable '$ENV_VAR_ARG' in an argument"
        ) in str(err.value)

    def test_alternate_command_executable(self, plugin_invoker):
        exec_args = plugin_invoker.exec_args(
            "extra",
            "args",
            command="alternate-exec",
            env={
                "ENV_VAR_ARG": "env-var-arg",
            },
        )

        assert exec_args[0].endswith("other-utility")
        assert exec_args[1:] == ["--option", "env-var-arg", "extra", "args"]

    @pytest.mark.parametrize(
        ("executable_str", "assert_fn"),
        (
            ("tap-test", lambda exe, _: exe == "tap-test"),
            ("./tap-test", lambda exe, name: exe.endswith(f"{name}/tap-test")),
            ("/apps/tap-test", lambda exe, _: exe == "/apps/tap-test"),
        ),
    )
    @pytest.mark.asyncio()
    async def test_expand_nonpip_command_exec_args(
        self,
        nonpip_plugin_invoker,
        session,
        executable_str,
        assert_fn,
    ):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        nonpip_plugin_invoker.plugin.executable = executable_str
        exec_args = nonpip_plugin_invoker.exec_args()

        assert assert_fn(exec_args[0], nonpip_plugin_invoker.project.root)

        await nonpip_plugin_invoker.prepare(session)
        env = nonpip_plugin_invoker.env()

        assert "VIRTUAL_ENV" not in env
        assert "PYTHONPATH" not in env



from __future__ import annotations

import logging
import re
import typing as t
from types import TracebackType

import pytest

from meltano.core.logging.formatters import console_log_formatter

if t.TYPE_CHECKING:
    from pathlib import Path

ANSI_RE = re.compile(r"\033\[[;?0-9]*[a-zA-Z]")


class FakeCode:
    """A fake code object that can be used to test log formatters."""

    def __init__(self, co_filename, co_name):
        self.co_filename = co_filename
        self.co_name = co_name


class FakeFrame:
    """A fake traceback frame that can be used to test log formatters."""

    def __init__(self, f_code, f_globals, f_locals=None):
        self.f_code = f_code
        self.f_globals = f_globals
        self.f_locals = f_locals or {}


class FakeTraceback:  # pragma: no cover
    """A fake traceback that can be used to test log formatters."""

    def __init__(self, frames, line_nums):
        if len(frames) != len(line_nums):
            raise ValueError("Ya messed up!")  # noqa: EM101
        self._frames = frames
        self._line_nums = line_nums
        self.tb_frame = frames[0]
        self.tb_lineno = line_nums[0]

    @property
    def tb_next(self):
        if len(self._frames) > 1:  # noqa: RET503
            return FakeTraceback(self._frames[1:], self._line_nums[1:])


class TestLogFormatters:
    @pytest.fixture()
    def record(self):
        return logging.LogRecord(
            name="test",
            level=logging.INFO,
            pathname="test",
            lineno=1,
            msg="test",
            args=None,
            exc_info=None,
        )

    @pytest.fixture()
    def record_with_exception(self, tmp_path: Path):
        path = tmp_path / "my_module.py"
        path.write_text("cause_an_error()\n")

        tb = t.cast(
            TracebackType,
            FakeTraceback(
                [
                    FakeFrame(
                        FakeCode(str(path), "test"),
                        {"__name__": "__main__"},
                        {"my_var": "my_value"},
                    ),
                ],
                [1],
            ),
        )

        return logging.LogRecord(
            name="test",
            level=logging.DEBUG,
            pathname="test",
            lineno=1,
            msg="test",
            args=None,
            exc_info=(
                Exception,
                Exception("test"),
                tb,
            ),
        )

    def test_console_log_formatter_colors(self, record):
        formatter = console_log_formatter(colors=True)
        assert ANSI_RE.match(formatter.format(record))

    def test_console_log_formatter_no_colors(self, record):
        formatter = console_log_formatter(colors=False)
        assert not ANSI_RE.match(formatter.format(record))

    def test_console_log_formatter_no_locals(self, record_with_exception):
        formatter = console_log_formatter(show_locals=False)
        output = formatter.format(record_with_exception)
        assert "locals" not in output
        assert "my_var = 'my_value'" not in output

    def test_console_log_formatter_show_locals(self, record_with_exception):
        formatter = console_log_formatter(show_locals=True)
        output = formatter.format(record_with_exception)
        assert "locals" in output
        assert "my_var = 'my_value'" in output



from __future__ import annotations

import asyncio
import json
import platform

import pytest
from mock import AsyncMock, mock

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.cli.utils import CliError
from meltano.core.job import Job, State
from meltano.core.plugin import PluginType
from meltano.core.plugin.singer import SingerTap
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.project_add_service import PluginAlreadyAddedException
from meltano.core.runner.dbt import DbtRunner
from meltano.core.runner.singer import SingerRunner


class LogEntry:
    def __init__(
        self,
        name: str | None = None,
        cmd_type: str | None = None,
        event: str | None = None,
        level: str | None = None,
        stdio: str | None = None,
    ):
        """Check whether a log entry is in a list of dicts.

        Args:
            name: The name to search for.
            cmd_type: The `cmd_type` to search for.
            event: Prefix of the event to search for.
            level: The level to search for.
            stdio: If not `None`, verify the stdio matches.
        """
        self.name = name
        self.cmd_type = cmd_type
        self.event = event
        self.level = level
        self.stdio = stdio

    def matches(self, lines: list[dict]) -> bool:
        """Find a matching log line in the provided list of log lines.

        It's important to note that the 'event' field check doesn't look for
        exact matches, and is doing a prefix search. This is because quite a
        few log lines have dynamic suffix segments.

        Args:
            lines: the log lines to check against

        Returns:
            True if a matching log line is found, else False
        """
        for line in lines:  # noqa: RET503
            matches = (
                line.get("name") == self.name
                and line.get("cmd_type") == self.cmd_type
                and line.get("event").startswith(self.event)
                and line.get("level") == self.level
            )

            if matches:
                return line.get("stdio") == self.stdio if self.stdio else True


def assert_lines(output, *lines):
    for line in lines:
        assert line in output


def exception_logged(result_output: str, exc: Exception) -> bool:
    """Small utility to search click result output for a specific exception.

    Args:
        result_output: The click result output string to search.
        exc: The exception to search for.

    Returns:
        bool: Whether the exception was found
    """
    seen_lines: list[dict] = []
    for line in result_output.splitlines():
        parsed_line = json.loads(line)
        seen_lines.append(parsed_line)

    for line in seen_lines:
        if line.get("exc_info") and repr(exc) in line.get("exc_info"):
            return True
    return False


def assert_log_lines(result_output: str, expected: list[LogEntry]):
    seen_lines: list[dict] = []
    for line in result_output.splitlines():
        parsed_line = json.loads(line)
        seen_lines.append(parsed_line)

    for entry in expected:
        assert entry.matches(seen_lines)


def failure_help_log_suffix(job_logs_file):
    return (
        "For more detailed log messages re-run the command using 'meltano "
        "--log-level=debug ...' CLI flag.\nNote that you can also check the "
        f"generated log file at '{job_logs_file}'.\nFor more information on "
        "debugging and logging: "
        "https://docs.meltano.com/reference/command-line-interface#debugging"
    )


@pytest.fixture(scope="class")
def tap_mock_transform(project_add_service):
    try:
        return project_add_service.add(PluginType.TRANSFORMS, "tap-mock-transform")
    except PluginAlreadyAddedException as err:
        return err.plugin


@pytest.fixture()
def process_mock_factory():
    def _factory(name):
        process_mock = mock.Mock()
        process_mock.name = name
        process_mock.wait = AsyncMock(return_value=0)
        process_mock.returncode = 0
        process_mock.stdin.wait_closed = AsyncMock(return_value=True)
        return process_mock

    return _factory


@pytest.fixture()
def tap_process(process_mock_factory, tap):
    tap = process_mock_factory(tap)
    tap.stdout.at_eof.side_effect = (False, False, False, True)
    tap.stdout.readline = AsyncMock(side_effect=(b"SCHEMA\n", b"RECORD\n", b"STATE\n"))
    tap.stderr.at_eof.side_effect = (False, False, False, True)
    tap.stderr.readline = AsyncMock(
        side_effect=(b"Starting\n", b"Running\n", b"Done\n"),
    )
    return tap


@pytest.fixture()
def target_process(process_mock_factory, target):
    target = process_mock_factory(target)

    # Have `target.wait` take 1s to make sure the tap always finishes before the target
    async def wait_mock():
        await asyncio.sleep(1)
        return target.wait.return_value

    target.wait.side_effect = wait_mock

    target.stdout.at_eof.side_effect = (False, False, False, True)
    target.stdout.readline = AsyncMock(
        side_effect=(b'{"line": 1}\n', b'{"line": 2}\n', b'{"line": 3}\n'),
    )
    target.stderr.at_eof.side_effect = (False, False, False, True)
    target.stderr.readline = AsyncMock(
        side_effect=(b"Starting\n", b"Running\n", b"Done\n"),
    )
    return target


@pytest.fixture()
def silent_dbt_process(process_mock_factory, dbt):
    dbt = process_mock_factory(dbt)
    dbt.stdout.at_eof.side_effect = (True, True)
    dbt.stderr.at_eof.side_effect = (True, True)
    return dbt


@pytest.fixture()
def dbt_process(process_mock_factory, dbt):
    dbt = process_mock_factory(dbt)
    dbt.stdout.at_eof.side_effect = (True,)
    dbt.stderr.at_eof.side_effect = (False, False, False, True)
    dbt.stderr.readline = AsyncMock(
        side_effect=(b"Starting\n", b"Running\n", b"Done\n"),
    )
    return dbt


class TestWindowsELT:
    @pytest.mark.skipif(
        platform.system() != "Windows",
        reason="Test is only for Windows",
    )
    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config")
    def test_elt_windows(
        self,
        cli_runner,
        tap,
        target,
    ):
        args = ["elt", tap.name, target.name]
        result = cli_runner.invoke(cli, args)
        assert result.exit_code == 1
        # Didn't use `exception_logged()` as `result.stderr` doesn't contain
        # the error for some reason
        assert (
            "ELT command not supported on Windows. Please use the run command "
            "as documented here: "
            "https://docs.meltano.com/reference/command-line-interface#run"
        ) in str(result.exception)


@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="ELT is not supported on Windows",
)
class TestCliEltScratchpadOne:
    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service,
        command: str,
    ):
        result = cli_runner.invoke(cli, [command])
        assert result.exit_code == 2

        state_id = f"pytest_test_{command}"
        args = [command, "--state-id", state_id, tap.name, target.name]

        # exit cleanly when everything is fine
        create_subprocess_exec = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock:
            asyncio_mock.create_subprocess_exec = create_subprocess_exec

            result = cli_runner.invoke(cli, args)
            assert_cli_runner(result)

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                    LogEntry(
                        None,
                        None,
                        "No state was found, complete import.",
                        "warning",
                    ),
                    LogEntry(
                        None,
                        None,
                        "Incremental state has been updated at",
                        "info",
                    ),
                    LogEntry("meltano", None, "Extract & load complete!", "info"),
                    LogEntry("meltano", None, "Transformation skipped.", "info"),
                ],
            )

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("tap-mock", "extractor", "Starting", "info"),
                    LogEntry("tap-mock", "extractor", "Running", "info"),
                    LogEntry("tap-mock", "extractor", "Done", "info"),
                    LogEntry("target-mock", "loader", "Starting", "info"),
                    LogEntry("target-mock", "loader", "Running", "info"),
                    LogEntry("target-mock", "loader", "Done", "info"),
                ],
            )

        job_logging_service.delete_all_logs(state_id)

        exc = Exception("This is a grave danger.")
        with mock.patch.object(SingerRunner, "run", side_effect=exc):
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert result.exception == exc

            lines = [
                LogEntry("meltano", None, "Running extract & load...", "info"),
                LogEntry(None, None, "This is a grave danger.", "error"),
            ]
            assert_log_lines(result.stderr, lines)
            assert exception_logged(result.stderr, exc)

            # ensure there is a log of this exception
            log = job_logging_service.get_latest_log(state_id).splitlines()
            assert "Traceback (most recent call last):" in log
            assert "Exception: This is a grave danger." in log

    @pytest.mark.flaky(reruns=3, reruns_delay=2)
    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt_debug_logging(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service,
        monkeypatch,
        command: str,
    ):
        state_id = f"pytest_test_{command}_debug"
        args = [command, "--state-id", state_id, tap.name, target.name]

        job_logging_service.delete_all_logs(state_id)

        create_subprocess_exec = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ), mock.patch("meltano.core.plugin_invoker.asyncio") as asyncio_mock:
            asyncio_mock.create_subprocess_exec = create_subprocess_exec

            monkeypatch.setenv("MELTANO_CLI_LOG_LEVEL", "debug")
            result = cli_runner.invoke(cli, args)
            assert_cli_runner(result)

            lines = [
                LogEntry("meltano", None, "Running extract & load...", "info"),
                LogEntry(
                    None,
                    None,
                    "Created configuration at",
                    "debug",
                ),  # followed by path
                LogEntry(
                    None,
                    None,
                    "Could not find tap.properties.json in",
                    "debug",
                ),  # followed by path
                LogEntry(
                    None,
                    None,
                    "Could not find state.json in",
                    "debug",
                ),  # followed by path
                LogEntry(
                    None,
                    None,
                    "Created configuration at",
                    "debug",
                ),  # followed by path
                LogEntry(None, None, "No state was found, complete import.", "warning"),
                LogEntry(
                    None,
                    None,
                    "Incremental state has been updated at",
                    "info",
                ),  # followed by timestamp
                LogEntry(
                    None,
                    None,
                    "Incremental state: {'line': 1}",
                    "debug",
                ),
                LogEntry(
                    None,
                    None,
                    "Incremental state: {'line': 2}",
                    "debug",
                ),
                LogEntry(
                    None,
                    None,
                    "Incremental state: {'line': 3}",
                    "debug",
                ),
                LogEntry("meltano", None, "Extract & load complete!", "info"),
                LogEntry("meltano", None, "Transformation skipped.", "info"),
                LogEntry("tap-mock", "extractor", "Starting", "info", "stderr"),
                LogEntry("tap-mock", "extractor", "Running", "info", "stderr"),
                LogEntry("tap-mock (out)", "extractor", "SCHEMA", "debug", "stdout"),
                LogEntry("tap-mock (out)", "extractor", "RECORD", "debug", "stdout"),
                LogEntry("tap-mock (out)", "extractor", "STATE", "debug", "stdout"),
                LogEntry("tap-mock", "extractor", "Done", "info", "stderr"),
                LogEntry("target-mock", "loader", "Starting", "info", "stderr"),
                LogEntry("target-mock", "loader", "Running", "info", "stderr"),
                LogEntry(
                    "target-mock (out)",
                    "loader",
                    '{"line": 1}',
                    "debug",
                    "stdout",
                ),
                LogEntry(
                    "target-mock (out)",
                    "loader",
                    '{"line": 2}',
                    "debug",
                    "stdout",
                ),
                LogEntry(
                    "target-mock (out)",
                    "loader",
                    '{"line": 3}',
                    "debug",
                    "stdout",
                ),
                LogEntry("target-mock", "loader", "Done", "info", "stderr"),
            ]

            assert_log_lines(result.stdout + result.stderr, lines)

            log = job_logging_service.get_latest_log(state_id)

            full_result = result.stdout + result.stderr

            # We expect a difference of 2 lines because the cli emits two log
            # lines not found in the log.
            # we already test the redirect handler in test_output_logger,
            # so we'll just verify that the # of lines matches.
            log_diff = 2 if command == "el" else 3
            assert len(log.splitlines()) == len(full_result.splitlines()) - log_diff
            # and just to be safe - check if these debug mode only strings show up
            assert "target-mock (out)" in log
            assert "tap-mock (out)" in log

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt_tap_failure(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service,
        command: str,
    ):
        state_id = f"pytest_test_{command}"
        args = [command, "--state-id", state_id, tap.name, target.name]

        tap_process.wait.return_value = 1
        tap_process.stderr.readline.side_effect = (
            b"Starting\n",
            b"Running\n",
            b"Failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(
            PluginInvoker,
            "invoke_async",
            new=invoke_async,
        ) as invoke_async:
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert "Extractor failed" in str(result.exception)
            job_logs_file = job_logging_service.get_all_logs(state_id)[0]

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                ],
            )

            assert exception_logged(
                result.stdout + result.stderr,
                CliError(
                    "ELT could not be completed: Extractor failed.\n"
                    + failure_help_log_suffix(job_logs_file),
                ),
            )

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("tap-mock", "extractor", "Starting", "info", "stderr"),
                    LogEntry("tap-mock", "extractor", "Running", "info", "stderr"),
                    LogEntry("tap-mock", "extractor", "Failure", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Starting", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Running", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Done", "info", "stderr"),
                ],
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt_target_failure_before_tap_finishes(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service,
        command: str,
    ):
        state_id = f"pytest_test_{command}"
        args = [command, "--state-id", state_id, tap.name, target.name]

        # Have `tap_process.wait` take 2s to make sure the target can fail
        # before tap finishes
        async def tap_wait_mock():
            await asyncio.sleep(2)
            return tap_process.wait.return_value

        tap_process.wait.side_effect = tap_wait_mock

        # Writing to target stdin will fail because (we'll pretend) it has already died
        target_process.stdin = mock.Mock(spec=asyncio.StreamWriter)
        target_process.stdin.write.side_effect = BrokenPipeError
        target_process.stdin.drain = AsyncMock(side_effect=ConnectionResetError)
        target_process.stdin.wait_closed = AsyncMock(return_value=True)

        # Have `target_process.wait` take 1s to make sure the
        # `stdin.write`/`drain` exceptions can be raised
        async def target_wait_mock():
            await asyncio.sleep(1)
            return 1

        target_process.wait.side_effect = target_wait_mock

        target_process.stderr.readline.side_effect = (
            b"Starting\n",
            b"Running\n",
            b"Failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(
            PluginInvoker,
            "invoke_async",
            new=invoke_async,
        ) as invoke_async:
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert "Loader failed" in str(result.exception)
            job_logs_file = job_logging_service.get_all_logs(state_id)[0]

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                ],
            )
            assert exception_logged(
                result.stdout + result.stderr,
                CliError(
                    "ELT could not be completed: Loader failed.\n"
                    + failure_help_log_suffix(job_logs_file),
                ),
            )

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("tap-mock", "extractor", "Starting", "info", "stderr"),
                    LogEntry("tap-mock", "extractor", "Running", "info", "stderr"),
                    LogEntry("tap-mock", "extractor", "Done", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Starting", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Running", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Failure", "info", "stderr"),
                ],
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt_target_failure_after_tap_finishes(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service,
        command: str,
    ):
        state_id = f"pytest_test_{command}"
        args = [command, "--state-id", state_id, tap.name, target.name]

        target_process.wait.return_value = 1
        target_process.stderr.readline.side_effect = (
            b"Starting\n",
            b"Running\n",
            b"Failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(
            PluginInvoker,
            "invoke_async",
            new=invoke_async,
        ) as invoke_async:
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert "Loader failed" in str(result.exception)
            job_logs_file = job_logging_service.get_all_logs(state_id)[0]

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                    LogEntry("meltano", None, "Loading failed", "error"),
                ],
            )
            assert exception_logged(
                result.stdout + result.stderr,
                CliError(
                    "ELT could not be completed: Loader failed.\n"
                    + failure_help_log_suffix(job_logs_file),
                ),
            )

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("tap-mock", "extractor", "Starting", "info", "stderr"),
                    LogEntry("tap-mock", "extractor", "Running", "info", "stderr"),
                    LogEntry("tap-mock", "extractor", "Done", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Starting", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Running", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Failure", "info", "stderr"),
                ],
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt_tap_and_target_failure(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service,
        command: str,
    ):
        state_id = f"pytest_test_{command}"
        args = [command, "--state-id", state_id, tap.name, target.name]

        tap_process.wait.return_value = 1
        tap_process.stderr.readline.side_effect = (
            b"Starting\n",
            b"Running\n",
            b"Failure\n",
        )

        target_process.wait.return_value = 1
        target_process.stderr.readline.side_effect = (
            b"Starting\n",
            b"Running\n",
            b"Failure\n",
        )

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(
            PluginInvoker,
            "invoke_async",
            new=invoke_async,
        ) as invoke_async:
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert "Extractor and loader failed" in str(result.exception)
            job_logs_file = job_logging_service.get_all_logs(state_id)[0]

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                    LogEntry("meltano", None, "Extraction failed", "error"),
                    LogEntry("meltano", None, "Loading failed", "error"),
                ],
            )

            assert exception_logged(
                result.stdout + result.stderr,
                CliError(
                    "ELT could not be completed: Extractor and loader failed.\n"
                    + failure_help_log_suffix(job_logs_file),
                ),
            )

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("tap-mock", "extractor", "Starting", "info", "stderr"),
                    LogEntry("tap-mock", "extractor", "Running", "info", "stderr"),
                    LogEntry("tap-mock", "extractor", "Failure", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Starting", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Running", "info", "stderr"),
                    LogEntry("target-mock", "loader", "Failure", "info", "stderr"),
                ],
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt_tap_line_length_limit_error(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        job_logging_service,
        command: str,
    ):
        state_id = f"pytest_test_{command}"
        args = [command, "--state-id", state_id, tap.name, target.name]

        # Raise a `ValueError` wrapping a `LimitOverrunError`, like
        # `StreamReader.readline` does:
        # https://github.com/python/cpython/blob/v3.8.7/Lib/asyncio/streams.py#L549
        try:  # noqa: WPS328
            raise asyncio.LimitOverrunError(
                "Separator is not found, and chunk exceed the limit",  # noqa: EM101
                0,
            )
        except asyncio.LimitOverrunError as err:
            try:  # noqa: WPS328, WPS505
                # `ValueError` needs to be raised from inside the except block
                # for `LimitOverrunError` so that `__context__` is set.
                raise ValueError(str(err))
            except ValueError as wrapper_err:
                tap_process.stdout.readline.side_effect = wrapper_err

        # Have `tap_process.wait` take 1s to make sure the `LimitOverrunError`
        # exception can be raised before tap finishes
        async def wait_mock():
            await asyncio.sleep(1)
            return tap_process.wait.return_value

        tap_process.wait.side_effect = wait_mock

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(
            PluginInvoker,
            "invoke_async",
            new=invoke_async,
        ) as invoke_async:
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert "Output line length limit exceeded" in str(result.exception)
            job_logs_file = job_logging_service.get_all_logs(state_id)[0]

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                    LogEntry(
                        None,
                        None,
                        (
                            "The extractor generated a message exceeding the "
                            "message size limit of 5.0MiB (half the buffer "
                            "size of 10.0MiB)."
                        ),
                        "error",
                    ),
                ],
            )

            assert exception_logged(
                result.stdout + result.stderr,
                CliError(
                    "ELT could not be completed: Output line length limit exceeded.\n"
                    + failure_help_log_suffix(job_logs_file),
                ),
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt_output_handler_error(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        command: str,
    ):
        state_id = f"pytest_test_{command}"
        args = [command, "--state-id", state_id, tap.name, target.name]

        exc = Exception("Failed to read from target stderr.")
        target_process.stderr.readline.side_effect = exc

        # Have `tap_process.wait` take 1s to make sure the exception can be
        # raised before tap finishes
        async def wait_mock():
            await asyncio.sleep(1)
            return tap_process.wait.return_value

        tap_process.wait.side_effect = wait_mock

        invoke_async = AsyncMock(side_effect=(tap_process, target_process))
        with mock.patch.object(
            PluginInvoker,
            "invoke_async",
            new=invoke_async,
        ) as invoke_async:
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert result.exception == exc

            assert_log_lines(
                result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                ],
            )

            assert exception_logged(
                result.stderr,
                Exception("Failed to read from target stderr."),
            )

    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_elt_already_running(self, cli_runner, tap, target, session, command: str):
        state_id = "already_running"
        args = [command, "--state-id", state_id, tap.name, target.name]

        existing_job = Job(job_name=state_id, state=State.RUNNING)
        existing_job.save(session)

        with mock.patch(
            "meltano.cli.elt.project_engine",
            return_value=(None, lambda: session),
        ):
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert f"Another '{state_id}' pipeline is already running" in str(
                result.exception,
            )

    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_dump_catalog(
        self,
        cli_runner,
        project,
        tap,
        target,
        command: str,
    ):
        catalog = {"streams": []}
        with project.root.joinpath("catalog.json").open("w") as catalog_file:
            json.dump(catalog, catalog_file)

        state_id = f"pytest_test_{command}"
        args = [
            command,
            "--state-id",
            state_id,
            tap.name,
            target.name,
            "--catalog",
            "catalog.json",
            "--dump",
            "catalog",
        ]

        result = cli_runner.invoke(cli, args)
        assert_cli_runner(result)

        assert json.loads(result.stdout) == catalog

    @pytest.mark.usefixtures("session")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_dump_state(
        self,
        cli_runner,
        project,
        tap,
        target,
        command: str,
    ):
        state = {"success": True}
        with project.root.joinpath("state.json").open("w") as state_file:
            json.dump(state, state_file)

        state_id = f"pytest_test_{command}"
        args = [
            command,
            "--state-id",
            state_id,
            tap.name,
            target.name,
            "--state",
            "state.json",
            "--dump",
            "state",
        ]

        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ):
            result = cli_runner.invoke(cli, args)
            assert_cli_runner(result)

            assert json.loads(result.stdout) == state

    @pytest.mark.usefixtures("project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_dump_extractor_config(
        self,
        cli_runner,
        tap,
        target,
        plugin_settings_service_factory,
        command: str,
    ):
        state_id = f"pytest_test_{command}"
        args = [
            command,
            "--state-id",
            state_id,
            tap.name,
            target.name,
            "--dump",
            "extractor-config",
        ]

        settings_service = plugin_settings_service_factory(tap)

        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ):
            result = cli_runner.invoke(cli, args)
            assert_cli_runner(result)

            assert json.loads(result.stdout) == settings_service.as_dict(
                extras=False,
                process=True,
            )

    @pytest.mark.usefixtures("project")
    @pytest.mark.parametrize("command", ("elt", "el"), ids=["elt", "el"])
    def test_dump_loader_config(
        self,
        cli_runner,
        tap,
        target,
        plugin_settings_service_factory,
        command: str,
    ):
        state_id = f"pytest_test_{command}"
        args = [
            command,
            "--state-id",
            state_id,
            tap.name,
            target.name,
            "--dump",
            "loader-config",
        ]

        settings_service = plugin_settings_service_factory(target)

        with mock.patch.object(SingerTap, "discover_catalog"), mock.patch.object(
            SingerTap,
            "apply_catalog_rules",
        ):
            result = cli_runner.invoke(cli, args)
            assert_cli_runner(result)

            assert json.loads(result.stdout) == settings_service.as_dict(
                extras=False,
                process=True,
            )


@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="ELT is not supported on Windows",
)
class TestCliEltScratchpadTwo:
    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "tap_mock_transform",
    )
    def test_elt_transform_run(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        silent_dbt_process,
        dbt_process,
    ):
        args = ["elt", tap.name, target.name, "--transform", "run"]

        invoke_async = AsyncMock(
            side_effect=(
                tap_process,
                target_process,
                silent_dbt_process,  # dbt clean
                silent_dbt_process,  # dbt deps
                dbt_process,  # dbt run
            ),
        )
        with mock.patch.object(
            PluginInvoker,
            "invoke_async",
            new=invoke_async,
        ) as invoke_async:
            result = cli_runner.invoke(cli, args)
            assert_cli_runner(result)

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                    LogEntry("meltano", None, "Extract & load complete!", "info"),
                    LogEntry("meltano", None, "Running transformation...", "info"),
                    LogEntry("meltano", None, "Transformation complete!", "info"),
                ],
            )

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("tap-mock", "extractor", "Starting", "info"),
                    LogEntry("tap-mock", "extractor", "Running", "info"),
                    LogEntry("tap-mock", "extractor", "Done", "info"),
                    LogEntry("target-mock", "loader", "Starting", "info"),
                    LogEntry("target-mock", "loader", "Running", "info"),
                    LogEntry("target-mock", "loader", "Done", "info"),
                    LogEntry("dbt", "transformer", "Starting", "info"),
                    LogEntry("dbt", "transformer", "Running", "info"),
                    LogEntry("dbt", "transformer", "Done", "info"),
                ],
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "tap_mock_transform",
    )
    def test_elt_transform_run_dbt_failure(
        self,
        cli_runner,
        tap,
        target,
        tap_process,
        target_process,
        silent_dbt_process,
        dbt_process,
        job_logging_service,
    ):
        state_id = "pytest_test_elt"
        args = [
            "elt",
            "--state-id",
            state_id,
            tap.name,
            target.name,
            "--transform",
            "run",
        ]

        dbt_process.wait.return_value = 1
        dbt_process.returncode = 1
        dbt_process.stderr.readline.side_effect = (
            b"Starting\n",
            b"Running\n",
            b"Failure\n",
        )

        invoke_async = AsyncMock(
            side_effect=(
                tap_process,
                target_process,
                silent_dbt_process,  # dbt clean
                silent_dbt_process,  # dbt deps
                dbt_process,  # dbt run
            ),
        )
        with mock.patch.object(
            PluginInvoker,
            "invoke_async",
            new=invoke_async,
        ) as invoke_async:
            result = cli_runner.invoke(cli, args)
            assert result.exit_code == 1
            assert "`dbt run` failed" in str(result.exception)
            job_logs_file = job_logging_service.get_all_logs(state_id)[0]

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Running extract & load...", "info"),
                    LogEntry("meltano", None, "Extract & load complete!", "info"),
                    LogEntry("meltano", None, "Running transformation...", "info"),
                    LogEntry("meltano", None, "Transformation failed", "error"),
                ],
            )
            assert exception_logged(
                result.stderr,
                CliError(
                    "ELT could not be completed: `dbt run` failed.\n"
                    + failure_help_log_suffix(job_logs_file),
                ),
            )

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("tap-mock", "extractor", "Starting", "info"),
                    LogEntry("tap-mock", "extractor", "Running", "info"),
                    LogEntry("tap-mock", "extractor", "Done", "info"),
                    LogEntry("target-mock", "loader", "Starting", "info"),
                    LogEntry("target-mock", "loader", "Running", "info"),
                    LogEntry("target-mock", "loader", "Done", "info"),
                    LogEntry("dbt", "transformer", "Starting", "info"),
                    LogEntry("dbt", "transformer", "Running", "info"),
                    LogEntry("dbt", "transformer", "Failure", "info"),
                ],
            )


@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="ELT is not supported on Windows",
)
class TestCliEltScratchpadThree:
    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures("use_test_log_config", "project", "dbt")
    def test_elt_transform_only(
        self,
        cli_runner,
        tap,
        target,
    ):
        args = ["elt", tap.name, target.name, "--transform", "only"]

        with mock.patch.object(DbtRunner, "run", new=AsyncMock()):
            result = cli_runner.invoke(cli, args)
            assert_cli_runner(result)

            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry(
                        None,
                        None,
                        "The `elt` command is deprecated in favor of `el`",
                        "warning",
                    ),
                    LogEntry("meltano", None, "Extract & load skipped.", "info"),
                    LogEntry("meltano", None, "Running transformation...", "info"),
                    LogEntry("meltano", None, "Transformation complete!", "info"),
                ],
            )

    @pytest.mark.backend("sqlite")
    @pytest.mark.usefixtures(
        "use_test_log_config",
        "project",
        "dbt",
        "tap_mock_transform",
    )
    def test_elt_transform_only_with_transform(
        self,
        cli_runner,
        tap,
        target,
    ):
        args = ["elt", tap.name, target.name, "--transform", "only"]

        with mock.patch.object(DbtRunner, "run", new=AsyncMock()):
            result = cli_runner.invoke(cli, args)
            assert_cli_runner(result)
            assert_log_lines(
                result.stdout + result.stderr,
                [
                    LogEntry("meltano", None, "Extract & load skipped.", "info"),
                    LogEntry("meltano", None, "Running transformation...", "info"),
                    LogEntry("meltano", None, "Transformation complete!", "info"),
                ],
            )



from __future__ import annotations

import pytest

from meltano.core.task_sets import InvalidTasksError, TaskSets, tasks_from_yaml_str


class TestTaskSets:
    def test_squash(self):
        tset = TaskSets(name="test", tasks=["tap target", "tap2 target2"])
        assert tset._as_args() == ["tap", "target", "tap2", "target2"]

        tset = TaskSets(
            name="test",
            tasks=[["tap target"], ["some:cmd"], ["tap2 target2"]],
        )
        assert tset._as_args() == ["tap", "target", "some:cmd", "tap2", "target2"]

        tset = TaskSets(
            name="test",
            tasks=[["tap target", "some:cmd"], ["tap2 target2"]],
        )
        assert tset._as_args(preserve_top_level=True) == [
            ["tap", "target", "some:cmd"],
            ["tap2", "target2"],
        ]

    def test_tasks_from_yaml_str(self):
        cases = [
            ("generic", "tap target", TaskSets(name="generic", tasks=["tap target"])),
            (
                "single-list",
                "[tap target]",
                TaskSets(name="single-list", tasks=["tap target"]),
            ),
            (
                "multiple-lists",
                "[tap target, tap2 target2]",
                TaskSets(name="multiple-lists", tasks=["tap target", "tap2 target2"]),
            ),
            (
                "multiple-lists-quoted",
                "['tap target', 'tap2 target2']",
                TaskSets(
                    name="multiple-lists-quoted",
                    tasks=["tap target", "tap2 target2"],
                ),
            ),
            (
                "nested-mixed-lists",
                "[['tap target', 'tap2 target2'], 'cmd1']",
                TaskSets(
                    name="nested-mixed-lists",
                    tasks=[["tap target", "tap2 target2"], "cmd1"],
                ),
            ),
        ]

        for name, task_str, expected in cases:
            tset = tasks_from_yaml_str(name, task_str)
            assert tset.name == expected.name
            assert tset.tasks == expected.tasks

        obvious_edge_cases = [
            ("bad-yaml", "['tap target'"),
            ("too-many-levels", "[[['tap target']]]"),
            ("non-string-list", "['tap target', 5"),
        ]

        for name, task_str in obvious_edge_cases:
            with pytest.raises(InvalidTasksError):
                tasks_from_yaml_str(name, task_str)



from __future__ import annotations

from meltano.core.block.parser import is_command_block


class TestParserUtils:
    def test_is_command_block(self, tap, dbt):
        assert not is_command_block(tap)
        assert is_command_block(dbt)



from __future__ import annotations

import json
import platform
import typing as t

import pytest
from mock import AsyncMock, mock

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.settings_service import REDACTED_VALUE, SettingValueStore

if t.TYPE_CHECKING:
    from pathlib import Path

    from meltano.core.project import Project


class TestCliConfig:
    @pytest.mark.usefixtures("project")
    def test_config(self, cli_runner, tap):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        result = cli_runner.invoke(cli, ["config", tap.name])
        assert_cli_runner(result)

        json_config = json.loads(result.stdout)
        assert json_config["test"] == "mock"

    @pytest.mark.usefixtures("project")
    def test_config_extras(self, cli_runner, tap):
        result = cli_runner.invoke(cli, ["config", "--extras", tap.name])
        assert_cli_runner(result)

        json_config = json.loads(result.stdout)
        assert "_select" in json_config

    @pytest.mark.usefixtures("project")
    def test_config_env(self, cli_runner, tap):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        result = cli_runner.invoke(cli, ["config", "--format=env", tap.name])
        assert_cli_runner(result)

        assert "TAP_MOCK_TEST='mock'" in result.stdout

    @pytest.mark.usefixtures("project")
    def test_config_meltano(self, cli_runner, engine_uri):
        result = cli_runner.invoke(cli, ["config", "meltano"])
        assert_cli_runner(result)

        json_config = json.loads(result.stdout)
        assert json_config["database_uri"] == engine_uri
        assert json_config["cli"]["log_level"] == "info"

    @pytest.mark.usefixtures("project")
    def test_config_meltano_set(self, cli_runner):
        result = cli_runner.invoke(
            cli,
            ["config", "meltano", "set", "cli.log_config", "log_config.yml"],
        )
        assert_cli_runner(result)
        assert (
            "Meltano setting 'cli.log_config' was set in `meltano.yml`: "
            "'log_config.yml'"
        ) in result.stdout

    @pytest.mark.usefixtures("project")
    def test_config_test(self, cli_runner, tap):
        mock_invoke = mock.Mock()
        mock_invoke.sterr.at_eof.side_effect = True
        mock_invoke.stdout.at_eof.side_effect = (False, True)
        mock_invoke.wait = AsyncMock(return_value=-1)
        mock_invoke.returncode = -1
        payload = json.dumps({"type": "RECORD"}).encode()
        mock_invoke.stdout.readline = AsyncMock(return_value=b"%b" % payload)

        with mock.patch(
            "meltano.core.plugin_invoker.PluginInvoker.invoke_async",
            return_value=mock_invoke,
        ) as mocked_invoke:
            result = cli_runner.invoke(cli, ["config", tap.name, "test"])
            assert mocked_invoke.assert_called_once
            assert_cli_runner(result)

            assert "Plugin configuration is valid" in result.stdout

    @pytest.mark.usefixtures("project")
    def test_config_meltano_test(self, cli_runner):
        result = cli_runner.invoke(cli, ["config", "meltano", "test"])

        assert result.exit_code == 1
        assert (
            str(result.exception)
            == "Testing of the Meltano project configuration is not supported"
        )

    @pytest.mark.usefixtures("project")
    def test_config_list_redacted(
        self,
        cli_runner,
        tap,
        session,
        plugin_settings_service_factory,
    ):
        plugin_settings_service = plugin_settings_service_factory(tap)
        plugin_settings_service.set(
            "secure",
            "thisisatest",
            store=SettingValueStore.DOTENV,
            session=session,
        )

        result = cli_runner.invoke(cli, ["config", tap.name, "list"])
        assert_cli_runner(result)

        assert (
            f"secure [env: TAP_MOCK_SECURE] current value: {REDACTED_VALUE} (from the "
            "TAP_MOCK_SECURE variable in `.env`)"
        ) in result.stdout

    @pytest.mark.usefixtures("project")
    def test_config_list_unsafe(
        self,
        cli_runner,
        tap,
        session,
        plugin_settings_service_factory,
    ):
        value = "thisisatest"

        plugin_settings_service = plugin_settings_service_factory(tap)
        plugin_settings_service.set(
            "secure",
            value,
            store=SettingValueStore.DOTENV,
            session=session,
        )

        result = cli_runner.invoke(cli, ["config", "--unsafe", tap.name, "list"])
        assert_cli_runner(result)

        assert (
            f"secure [env: TAP_MOCK_SECURE] current value: '{value}' (from the "
            "TAP_MOCK_SECURE variable in `.env`)"
        ) in result.stdout


class TestCliConfigSet:
    @pytest.mark.usefixtures("project")
    def test_config_set_redacted(self, cli_runner, tap):
        result = cli_runner.invoke(
            cli,
            ["config", tap.name, "set", "secure", "thisisatest"],
        )
        assert_cli_runner(result)

        assert (
            f"Extractor '{tap.name}' setting 'secure' was set in `.env`: "
            + REDACTED_VALUE
        ) in result.stdout

    @pytest.mark.usefixtures("project")
    def test_config_set_unsafe(self, cli_runner, tap):
        value = "thisisatest"

        result = cli_runner.invoke(
            cli,
            ["config", "--unsafe", tap.name, "set", "secure", value],
        )
        assert_cli_runner(result)

        assert (
            f"Extractor '{tap.name}' setting 'secure' was set in `.env`: '{value}'"
        ) in result.stdout

    @pytest.mark.usefixtures("project")
    def test_config_set_from_file(self, cli_runner, tap, tmp_path: Path):
        result = cli_runner.invoke(
            cli,
            ["config", tap.name, "set", "private_key", "--from-file", "-"],
            input="content-from-stdin",
        )
        assert_cli_runner(result)

        assert (
            f"Extractor '{tap.name}' setting 'private_key' was set in `meltano.yml`: "
            "'content-from-stdin'"
        ) in result.stdout

        filepath = tmp_path.joinpath("file.txt")
        filepath.write_text("content-from-file")

        result = cli_runner.invoke(
            cli,
            ["config", tap.name, "set", "private_key", "--from-file", filepath],
        )
        assert_cli_runner(result)

        assert (
            f"Extractor '{tap.name}' setting 'private_key' was set in `meltano.yml`: "
            "'content-from-file'"
        ) in result.stdout

    @pytest.mark.usefixtures("tap")
    def test_environments_order_of_precedence(self, project: Project, cli_runner):
        # set base config in `meltano.yml`
        result = cli_runner.invoke(
            cli,
            ["--no-environment", "config", "tap-mock", "set", "test", "base-mock"],
        )
        assert_cli_runner(result)
        base_tap_config = next(
            (
                tap
                for tap in project.meltano["plugins"]["extractors"]
                if tap["name"] == "tap-mock"
            ),
            {},
        )
        assert base_tap_config["config"]["test"] == "base-mock"

        # set base config in `meltano.yml` -- ignore default environment
        result = cli_runner.invoke(
            cli,
            ["config", "tap-mock", "set", "test", "base-mock-no-default"],
        )
        assert_cli_runner(result)
        base_tap_config = next(
            (
                tap
                for tap in project.meltano["plugins"]["extractors"]
                if tap["name"] == "tap-mock"
            ),
            {},
        )
        assert base_tap_config["config"]["test"] == "base-mock-no-default"

        # set dev environment config in `meltano.yml`
        result = cli_runner.invoke(
            cli,
            ["--environment=dev", "config", "tap-mock", "set", "test", "dev-mock"],
        )
        assert_cli_runner(result)
        dev_env = next(
            (env for env in project.meltano["environments"] if env["name"] == "dev"),
            {},
        )
        tap_config = next(
            (
                tap
                for tap in dev_env["config"]["plugins"]["extractors"]
                if tap["name"] == "tap-mock"
            ),
            {},
        )
        assert tap_config["config"]["test"] == "dev-mock"



from __future__ import annotations

import platform
from datetime import datetime

import mock
import pytest

from meltano.core.plugin import PluginType
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.project_plugins_service import PluginAlreadyAddedException
from meltano.core.schedule_service import (
    BadCronError,
    Schedule,
    ScheduleAlreadyExistsError,
    ScheduleDoesNotExistError,
    ScheduleNotFoundError,
    SettingMissingError,
)


@pytest.fixture(scope="session")
def create_elt_schedule():
    def make(name, **kwargs):
        attrs = {
            "extractor": "tap-mock",
            "loader": "target-mock",
            "transform": "run",
            "interval": "@daily",
            "start_date": None,
            "env": {},
        }

        attrs.update(kwargs)
        return Schedule(name=name, **attrs)

    return make


@pytest.fixture(scope="session")
def create_job_schedule():
    def make(name, **kwargs):
        attrs = {
            "job": "job-mock",
            "interval": "@daily",
            "start_date": None,
            "env": {},
        }

        attrs.update(kwargs)
        return Schedule(name=name, **attrs)

    return make


@pytest.fixture(scope="class")
def custom_tap(project):
    tap = ProjectPlugin(
        PluginType.EXTRACTORS,
        name="tap-custom",
        namespace="tap_custom",
    )
    try:
        return project.plugins.add_to_file(tap)
    except PluginAlreadyAddedException as err:
        return err.plugin


class TestScheduleService:
    @pytest.fixture()
    def subject(self, schedule_service):
        return schedule_service

    @pytest.mark.order(0)
    def test_add_schedules(self, subject, create_elt_schedule, create_job_schedule):
        intervals = [
            "@once",
            "@manual",
            "@none",
            "@hourly",
            "@daily",
            "@weekly",
            "@monthly",
            "@yearly",
        ]

        job_schedules = [
            create_job_schedule(f"job_schedule_{interval[1:]}", interval=interval)
            for interval in intervals
        ]
        elt_schedules = [
            create_elt_schedule(f"elt_schedule_{interval[1:]}", interval=interval)
            for interval in intervals
        ]

        all_schedules = job_schedules + elt_schedules

        for schedule in all_schedules:
            subject.add_schedule(schedule)

        assert subject.schedules() == all_schedules

        # but name must be unique
        with pytest.raises(ScheduleAlreadyExistsError):
            subject.add_schedule(all_schedules[0])

        with pytest.raises(BadCronError) as excinfo:
            subject.add_schedule(create_elt_schedule("bad-cron", interval="bad_cron"))

        assert "bad_cron" in str(excinfo.value)
        assert excinfo.value.reason == "Invalid Cron expression or alias: 'bad_cron'"
        assert excinfo.value.instruction == "Please use a valid cron expression"

    def test_remove_schedule(self, subject):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        schedules = list(subject.schedules())
        schedules_count = len(schedules)

        idx = 3
        target_schedule = schedules[idx]
        target_name = target_schedule.name

        subject.remove_schedule(target_name)

        # make sure one has been removed
        schedules = list(subject.schedules())
        assert len(schedules) == schedules_count - 1
        assert target_schedule not in schedules

        # schedule name must exist to be removed
        with pytest.raises(ScheduleDoesNotExistError):
            subject.remove_schedule(target_name)

    def test_schedule_update(self, subject):
        schedule = subject.schedules()[0]

        yearly_intervals = sum(sbj.interval == "@yearly" for sbj in subject.schedules())

        schedule.interval = "@yearly"
        subject.update_schedule(schedule)

        # there should be one more schedule with the "@yearly" interval after set
        assert (
            sum(sbj.interval == "@yearly" for sbj in subject.schedules())
            == yearly_intervals + 1
        )

        # it should be the first element
        assert subject.schedules()[0].interval == "@yearly"

        # it should be a copy of the original element
        assert schedule is not subject.schedules()[0]

        # it must exists
        schedule.name = "llamasareverynice"
        with pytest.raises(ScheduleDoesNotExistError):
            subject.update_schedule(schedule)

    def test_schedule_start_date(
        self,
        subject,
        session,
        tap,
        target,
        plugin_settings_service_factory,
    ):
        # curry the `add_elt` method to remove some arguments
        def add_elt(name, start_date):
            return subject.add_elt(  # noqa: E731
                session,
                name,
                tap.name,
                target.name,
                "run",
                "@daily",
                start_date=start_date,
            )

        mock_date = datetime(2002, 1, 1)  # noqa: WPS432

        # when a start_date is set, the schedule should use it
        schedule = add_elt("with_start_date", mock_date)
        assert schedule.start_date == mock_date

        # or use the start_date in the extractor configuration
        plugin_settings_service = plugin_settings_service_factory(tap)
        plugin_settings_service.set("start_date", mock_date, session=session)
        schedule = add_elt("with_default_start_date", None)
        assert schedule.start_date == mock_date

        # or default to `utcnow()` if the plugin exposes no config
        with mock.patch(
            "meltano.core.schedule_service.PluginSettingsService.get",
            side_effect=SettingMissingError("start_date"),
        ):
            schedule = add_elt("with_no_start_date", None)
            assert schedule.start_date

    def test_run_elt_schedule(self, subject, session, tap, target):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        schedule = subject.add_elt(
            session,
            "tap-to-target",
            tap.name,
            target.name,
            "skip",
            "@daily",
            TAP_MOCK_TEST="overridden",
        )

        # It fails because tap and target are not actually installed
        process = subject.run(schedule)
        assert process.returncode == 1

        process_mock = mock.Mock(returncode=0)
        with mock.patch(
            "meltano.core.schedule_service.MeltanoInvoker.invoke",
            return_value=process_mock,
        ) as invoke_mock:
            process = subject.run(
                schedule,
                "--dump=config",
                env={"TAP_MOCK_SECURE": "overridden"},
            )
            assert process.returncode == 0

            invoke_mock.assert_called_once_with(
                [
                    "elt",
                    tap.name,
                    target.name,
                    f"--transform={schedule.transform}",
                    f"--state-id={schedule.name}",
                    "--dump=config",
                ],
                env={"TAP_MOCK_TEST": "overridden", "TAP_MOCK_SECURE": "overridden"},
            )

    @pytest.mark.usefixtures("session", "tap", "target")
    def test_run_job_schedule(self, subject):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        schedule = subject.add(
            "mock-job-schedule",
            "mock-job",
            "@daily",
            MOCK_ENV_FROM_SCHED="env_var_value_from_schedule_def",
        )

        # It fails because mock-job is not a valid job
        assert subject.run(schedule).returncode == 1

        process_mock = mock.Mock(returncode=0)
        with mock.patch(
            "meltano.core.schedule_service.MeltanoInvoker.invoke",
            return_value=process_mock,
        ) as invoke_mock:
            process = subject.run(
                schedule,
                "--dry-run",
                env={"MOCK_ENV_ENTRY": "athing"},
            )
            assert process.returncode == 0

            invoke_mock.assert_called_once_with(
                [
                    "run",
                    "--dry-run",
                    schedule.job,
                ],
                env={
                    "MOCK_ENV_ENTRY": "athing",
                    "MOCK_ENV_FROM_SCHED": "env_var_value_from_schedule_def",
                },
            )

    def test_find_namespace_schedule(self, subject, tap, create_elt_schedule):
        schedule = create_elt_schedule(tap.name)
        subject.add_schedule(schedule)
        found_schedule = subject.find_namespace_schedule(tap.namespace)
        assert found_schedule.extractor == tap.name

    @pytest.mark.usefixtures("create_elt_schedule")
    def test_find_namespace_schedule_custom_extractor(
        self,
        subject,
        custom_tap,
    ):
        schedule = Schedule(
            name="tap-custom",
            extractor="tap-custom",
            interval="@daily",
        )
        subject.add_schedule(schedule)
        found_schedule = subject.find_namespace_schedule(custom_tap.namespace)
        assert found_schedule.extractor == custom_tap.name

    def test_find_namespace_schedule_not_found(self, subject):
        with pytest.raises(ScheduleNotFoundError):
            subject.find_namespace_schedule("no-such-namespace")



from __future__ import annotations

import logging
import os
import typing as t
from pathlib import Path

import pytest
from click.testing import CliRunner

from fixtures.utils import cd, tmp_project
from meltano.core.project_files import ProjectFiles

if t.TYPE_CHECKING:
    from click import Command
    from click.testing import Result

    from fixtures.docker import SnowplowMicro


current_dir = Path(__file__).parent
plugins_dir = current_dir / "plugins"


class MeltanoCliRunner(CliRunner):
    def __init__(self, *args, snowplow: SnowplowMicro | None = None, **kwargs):
        """Initialize the `MeltanoCliRunner`."""
        self.snowplow = snowplow
        super().__init__(*args, **kwargs)

    def invoke(self, cli: Command, *args, **kwargs) -> Result:
        results = super().invoke(cli, *args, **kwargs)
        if self.snowplow:  # pragma: no cover
            assert self.snowplow.all()["bad"] == 0  # pragma: no cover
            assert not self.snowplow.bad()  # pragma: no cover
        return results


@pytest.fixture()
def cli_runner(pushd, snowplow_optional: SnowplowMicro | None):
    pushd(os.getcwd())  # Ensure we return to the CWD after the test
    root_logger = logging.getLogger()
    log_level = root_logger.level
    try:
        yield MeltanoCliRunner(mix_stderr=False, snowplow=snowplow_optional)
    finally:
        root_logger.setLevel(log_level)


@pytest.fixture(scope="class")
def large_config_project(
    compatible_copy_tree,
    tmp_path_factory: pytest.TempPathFactory,
):
    with cd(tmp_path_factory.mktemp("meltano-large-config-project")), tmp_project(
        "large_config_project",
        current_dir / "large_config_project",
        compatible_copy_tree,
    ) as project:
        yield project


@pytest.fixture(scope="class")
def project_files_cli(compatible_copy_tree, tmp_path_factory: pytest.TempPathFactory):
    with cd(tmp_path_factory.mktemp("meltano-project-files-cli")), tmp_project(
        "a_multifile_meltano_project_cli",
        current_dir / "multifile_project",
        compatible_copy_tree,
    ) as project:
        yield ProjectFiles(root=project.root, meltano_file_path=project.meltanofile)



from __future__ import annotations

import json
import platform

import pytest

from meltano.core.state_service import InvalidJobStateError
from meltano.core.utils import merge


class TestStateService:
    def test_validate_state(self, state_service):
        with pytest.raises(InvalidJobStateError):
            state_service.validate_state(
                json.loads('{"root key not singer_state": {}}'),
            )
        assert (
            state_service.validate_state(
                json.loads(
                    '{"singer_state": {"bookmarks": {"mock-stream": "mock-value"}}}',
                ),
            )
            is None
        )

    def test_get_state(self, state_service, state_ids_with_expected_states):
        for state_id, expected_state in state_ids_with_expected_states:
            assert state_service.get_state(state_id) == expected_state

    def test_list_state(self, state_service, state_ids_with_expected_states):
        assert state_service.list_state() == dict(state_ids_with_expected_states)

    def test_add_state(self, state_service, payloads):
        mock_state_id = "nonexistent"
        state_service.add_state(
            mock_state_id,
            json.dumps(payloads.mock_state_payloads[0]),
        )
        assert state_service.get_state(mock_state_id) == payloads.mock_state_payloads[0]

    @pytest.mark.usefixtures("job_history_session")
    def test_set_state(self, jobs, payloads, state_service):
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        for job in jobs:
            for state in payloads.mock_state_payloads:
                state_service.set_state(job.job_name, json.dumps(state))
                assert state_service.get_state(job.job_name) == state

    @pytest.mark.usefixtures("job_history_session")
    def test_clear_state(self, jobs, payloads, state_service):
        for job in jobs:
            state_service.clear_state(job.job_name)
            assert state_service.get_state(job.job_name) == payloads.mock_empty_payload

    @pytest.mark.usefixtures("job_history_session")
    def test_merge_state(self, jobs, state_service):
        job_pairs = [(jobs[idx], jobs[idx + 1]) for idx in range(0, len(jobs) - 1, 2)]
        for job_src, job_dst in job_pairs:
            state_src = state_service.get_state(job_src.job_name)
            state_dst = state_service.get_state(job_dst.job_name)
            merged_dst = merge(state_src, state_dst)
            state_service.merge_state(job_src.job_name, job_dst.job_name)
            assert merged_dst == state_service.get_state(job_dst.job_name)

    def test_copy(self, state_ids, state_service):
        state_id_pairs = [
            (state_ids[idx], state_ids[idx + 1])
            for idx in range(0, len(state_ids) - 1, 2)
        ]
        for state_id_src, state_id_dst in state_id_pairs:
            state_src = state_service.get_state(state_id_src)
            state_service.copy_state(state_id_src, state_id_dst)
            assert state_service.get_state(state_id_dst) == state_src

    def test_move(self, state_ids, state_service):
        state_id_pairs = [
            (state_ids[idx], state_ids[idx + 1])
            for idx in range(0, len(state_ids) - 1, 2)
        ]
        for state_id_src, state_id_dst in state_id_pairs:
            state_src = state_service.get_state(state_id_src)
            state_service.move_state(state_id_src, state_id_dst)
            assert not state_service.get_state(state_id_src)
            assert state_service.get_state(state_id_dst) == state_src



"""Meltano upgrade service."""

from __future__ import annotations

import os
import subprocess
import sys
import typing as t

import click

import meltano
from meltano.cli.utils import PluginInstallReason, install_plugins
from meltano.core.error import MeltanoError
from meltano.core.project_plugins_service import PluginType
from meltano.core.state_service import StateService
from meltano.core.state_store.filesystem import CloudStateStoreManager

if t.TYPE_CHECKING:
    from sqlalchemy.engine import Engine

    from meltano.core.project import Project


class UpgradeError(Exception):
    """The Meltano upgrade fails."""


class AutomaticPackageUpgradeError(Exception):
    """An automatic upgrade of Meltano fails."""

    def __init__(self, reason: str, instructions: str):
        """Initialize the `AutomaticPackageUpgradeError`.

        Args:
            reason: The reason the exception occured.
            instructions: Instructions for how to manually resolve the exception.
        """
        self.reason = reason
        self.instructions = instructions


class UpgradeService:  # noqa: WPS214
    """Meltano upgrade service."""

    def __init__(self, engine: Engine, project: Project):
        """Initialize the Meltano upgrade service.

        Args:
            engine: The SQLAlchemy engine to be used for the upgrade.
            project: The Meltano project.
        """
        self.project = project
        self.engine = engine

    def _upgrade_package(self, pip_url: str | None, force: bool) -> bool:
        fail_reason = None
        instructions = ""

        meltano_file_path = "/src/meltano/__init__.py"
        editable = meltano.__file__.endswith(meltano_file_path)
        if editable and not force:
            meltano_dir = meltano.__file__[: -len(meltano_file_path)]
            fail_reason = "it is installed from source"
            instructions = f"navigate to `{meltano_dir}` and run `git pull`"

        elif os.path.exists("/.dockerenv"):
            fail_reason = "it is installed inside Docker"
            instructions = (
                "pull the latest Docker image using "
                "`docker pull meltano/meltano` and recreate any containers "
                "you may have created"
            )

        elif os.getenv("NOX_CURRENT_SESSION") == "tests":
            fail_reason = "it is installed inside a Nox test session"
            instructions = ""

        if fail_reason:
            raise AutomaticPackageUpgradeError(
                reason=fail_reason,
                instructions=instructions,
            )

        pip_url = pip_url or "meltano"
        run = subprocess.run(
            [sys.executable, "-m", "pip", "install", "--upgrade", pip_url],
            stderr=subprocess.PIPE,
            text=True,
        )

        if run.returncode != 0:
            raise UpgradeError("Failed to upgrade `meltano`.", run)  # noqa: EM101

        return True

    def upgrade_package(self, pip_url: str | None = None, force: bool = False) -> bool:
        """Upgrade the Meltano package.

        Args:
            pip_url: The pip URL to use when upgrading the Meltano package.
            force: Whether editable installations from source should be overwritten.

        Returns:
            Whether the upgrade was successful.
        """
        click.secho("Upgrading `meltano` package...", fg="blue")

        try:
            self._upgrade_package(pip_url, force)
        except AutomaticPackageUpgradeError as err:
            msg = click.style(
                "The `meltano` package could not be upgraded automatically",
                fg="red",
            )
            click.echo(f"{msg} because {err.reason}.")
            if err.instructions:
                click.echo(f"To upgrade manually, {err.instructions}.")
            return False

        click.echo("The `meltano` package has been upgraded.")
        click.echo()
        return True

    def update_files(self):
        """Update the files managed by Meltano inside the current project.

        Raises:
            MeltanoError: Failed to upgrade plugins.
        """
        click.secho("Updating files managed by plugins...", fg="blue")

        file_plugins = self.project.plugins.get_plugins_of_type(PluginType.FILES)
        if not file_plugins:
            click.echo("Nothing to update")
            return

        success = install_plugins(
            self.project,
            file_plugins,
            reason=PluginInstallReason.UPGRADE,
        )
        if not success:
            raise MeltanoError("Failed to upgrade plugin(s)")  # noqa: EM101

    def migrate_database(self):
        """Migrate the Meltano database.

        Raises:
            UpgradeError: The migration failed.
        """
        click.secho("Applying migrations to system database...", fg="blue")

        from meltano.core.migration_service import MigrationError, MigrationService

        try:
            migration_service = MigrationService(self.engine)
            migration_service.upgrade()
        except MigrationError as err:
            raise UpgradeError(str(err)) from err

    def migrate_state(self):
        """Move cloud state files to deduplicated prefix paths.

        See: https://github.com/meltano/meltano/issues/7938
        """
        state_service = StateService(project=self.project)
        manager = state_service.state_store_manager
        if isinstance(manager, CloudStateStoreManager):
            click.secho("Applying migrations to project state...", fg="blue")
            for filepath in state_service.state_store_manager.list_all_files():
                parts = filepath.split(manager.delimiter)
                if (
                    parts[-1] == "state.json"
                    and filepath.count(manager.prefix.strip(manager.delimiter)) > 1
                ):
                    duplicated_substr = manager.delimiter.join(
                        [
                            manager.prefix.strip(manager.delimiter),
                            manager.prefix.strip(manager.delimiter),
                        ],
                    )
                    new_path = filepath.replace(duplicated_substr, manager.prefix)
                    new_path = new_path.replace(
                        manager.delimiter * 2,
                        manager.delimiter,
                    )
                    manager.copy_file(filepath, new_path)
                    click.secho(f"Copied state from {filepath} to {new_path}")

    def upgrade(self, skip_package: bool = False, **kwargs):  # noqa: WPS213
        """Upgrade Meltano.

        Note: this is not actually called as part of the `meltano upgrade` command
        but is useful for testing and debugging upgrade logic.

        Args:
            skip_package: Whether the Meltano package should be upgraded.
            kwargs: Keyword arguments for `UpgradeService.upgrade_package`.
        """
        package_upgraded = False
        if not skip_package:
            package_upgraded = self.upgrade_package(**kwargs)

            if not package_upgraded:
                click.echo(
                    "Then, run `meltano upgrade --skip-package` to upgrade "
                    "your project based on the latest version.",
                )
                return

            click.echo()

        self.update_files()
        click.echo()
        self.migrate_database()
        click.echo()
        self.migrate_state()
        click.echo()
        click.secho(
            "Meltano and your Meltano project have been upgraded!"
            if package_upgraded
            else "Your Meltano project has been upgraded!",
            fg="green",
        )



from __future__ import annotations

import os
import shutil
import typing as t
from functools import partial

import pytest

from fixtures.utils import cd

if t.TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture(scope="class")
def compatible_copy_tree():
    """Copy files recursively from source to destination, ignoring existing dirs."""

    def _compatible_copy_tree(source: Path, destination: Path):
        """Copy files recursively from source to destination, ignoring existing dirs."""
        shutil.copytree(source, destination, dirs_exist_ok=True)

    return _compatible_copy_tree


@pytest.fixture()
def function_scoped_test_dir(tmp_path_factory) -> Path:
    tmp_path = tmp_path_factory.mktemp("meltano_root")
    cwd = os.getcwd()
    try:
        os.chdir(tmp_path)
        yield tmp_path
    finally:
        os.chdir(cwd)


@pytest.fixture()
def empty_meltano_yml_dir(tmp_path):
    with cd(tmp_path):
        (tmp_path / "meltano.yml").touch()
        return tmp_path


@pytest.fixture()
def pushd(request):
    def _pushd(path):
        popd = partial(os.chdir, os.getcwd())
        request.addfinalizer(popd)
        os.chdir(path)

        return popd

    return _pushd


@pytest.mark.meta()
def test_pushd(tmp_path, pushd):
    os.makedirs(tmp_path / "a")
    os.makedirs(tmp_path / "a" / "b")

    pushd(tmp_path / "a")
    assert os.getcwd() == str(tmp_path / "a")

    popd = pushd("b")
    assert os.getcwd() == str(tmp_path / "a" / "b")

    popd()
    assert os.getcwd() == str(tmp_path / "a")



"""SingerTarget and supporting classes."""

from __future__ import annotations

import json
import logging
import typing as t
from datetime import datetime

from meltano.core.behavior.hookable import hook
from meltano.core.job import Job, Payload
from meltano.core.setting_definition import SettingDefinition
from meltano.core.state_service import SINGER_STATE_KEY, StateService

from . import PluginType, SingerPlugin

if t.TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from meltano.core.plugin_invoker import PluginInvoker

logger = logging.getLogger(__name__)


class BookmarkWriter:
    """A basic bookmark writer suitable for use as an output handler.

    Has a writelines method to support ingesting and persisting state messages.
    """

    def __init__(
        self,
        job: Job,
        session: Session,
        payload_flag: int = Payload.STATE,
        state_service: StateService | None = None,
    ):
        """Initialize the `BookmarkWriter`.

        Args:
            job: meltano elt job associated with this invocation and whose
                state will be updated.
            session: SQLAlchemy session/engine object to be used to update state.
            payload_flag: A payload flag.
            state_service: `StateService` to use for bookmarking state.
        """
        self.job = job
        self.session = session
        self.state_service = state_service or StateService(session=self.session)
        self.payload_flag = payload_flag

    def writeline(self, line: str):
        """Persist a state entry.

        Args:
            line: raw json state line to decode/store
        """
        if self.job is None:
            logger.info(
                "Running outside a Job context: "
                "incremental state could not be updated.",
            )
            return

        new_state = {}
        try:
            new_state = json.loads(line)
        except Exception:
            logger.warning(
                "Received state is invalid, incremental state has not been updated",
            )

        job = self.job
        job.payload[SINGER_STATE_KEY] = new_state
        job.payload_flags |= self.payload_flag
        try:
            job.save(self.session)
            self.state_service.add_state(
                job,
                json.dumps(job.payload),
                job.payload_flags,
            )
        except Exception:
            logger.warning(
                "Unable to persist state, or received state is invalid, "
                "incremental state has not been updated",
            )
        else:
            logger.info(f"Incremental state has been updated at {datetime.utcnow()}.")  # noqa: G004
            logger.debug(f"Incremental state: {new_state}")  # noqa: G004


class SingerTarget(SingerPlugin):
    """A plugin for singer targets."""

    __plugin_type__ = PluginType.LOADERS

    EXTRA_SETTINGS = [
        SettingDefinition(name="_dialect", value="$MELTANO_LOADER_NAMESPACE"),
    ]

    def exec_args(self, plugin_invoker):
        """Get command-line args to pass to the plugin.

        Args:
            plugin_invoker: PluginInvoker running this target.

        Returns:
            Command-line args for target
        """
        return ["--config", plugin_invoker.files["config"]]

    @property
    def config_files(self):
        """Get config files for this target.

        Returns:
            The config_files for this target.
        """
        return {"config": f"target.{self.instance_uuid}.config.json"}

    @property
    def output_files(self):
        """Get output files for this target.

        Returns:
            The output files for this target.
        """
        return {"state": "new_state.json"}

    @hook("before_invoke")
    async def setup_bookmark_writer_hook(
        self,
        plugin_invoker: PluginInvoker,
        exec_args: list[str],
    ):
        """Before invoke hook to trigger setting up the bookmark writer for this target.

        Args:
            plugin_invoker: The invocation handler of the plugin instance.
            exec_args: List of subcommand/args that we where invoked with.
        """
        if exec_args is None:
            exec_args = []

        if "--discover" in exec_args or "--help" in exec_args:
            return

        self.setup_bookmark_writer(plugin_invoker)

    def setup_bookmark_writer(self, plugin_invoker: PluginInvoker):
        """Configure the bookmark writer.

        If running in a pipeline context, we configure the bookmark writer as
        an additional output handler on the invoker.

        This leverages calling back to `PluginInvokers.add_output_handler` to
        attach an additional output handler (the `BookmarkWriter`) to handle
        persisting state messages.

        Args:
            plugin_invoker: The invocation handler whose `add_out_handler` method
                will be called to attach the bookmark writer as an additional
                output handler.
        """
        elt_context = plugin_invoker.context
        if not elt_context or not elt_context.job or not elt_context.session:
            return

        incomplete_state = (
            elt_context.full_refresh and elt_context.select_filter
        ) or elt_context.merge_state
        payload_flag = Payload.INCOMPLETE_STATE if incomplete_state else Payload.STATE

        plugin_invoker.add_output_handler(
            plugin_invoker.StdioSource.STDOUT,
            BookmarkWriter(elt_context.job, elt_context.session, payload_flag),
        )
